import math
from typing import Any

import numpy as np
from numpy.random import Generator
from scipy.optimize import root_scalar

from cratermaker.components.projectile import Projectile
from cratermaker.components.scaling import Scaling
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike
from cratermaker.utils import montecarlo_utils as mc
from cratermaker.utils.general_utils import (
    _create_catalogue,
    _set_properties,
    format_large_units,
    parameter,
)


@Scaling.register("montecarlo")
class MonteCarloScaling(Scaling):
    """
    An operations class for computing the scaling relationships between projectiles and craters.

    This class encapsulates the logic for converting between projectile properties and crater properties, as well as determining crater morphology based on size and target properties. This implements the scaling laws similar to those in Richardson (2009) that were implemented in CTEM. However, unlike in CTEM, we apply monte carlo methods to the scaling laws to account for the uncertainty in the scaling laws. We have also included updated simple-to-complex transition diameter values from Schenk et al. (2021).

    Parameters
    ----------
    target : Target | str, default="Moon"
        The target body for the impact. Can be a Target object or a string representing the target name.
    projectile : Projectile | str, default="asteroids"
        The projectile model for the impact. Can be an Projectile object or a string representing the projectile name.
    material : str or None
        Name of the target material composition of the target body to look up from the built-in catalogue. Options include "water", "sand", "dry soil", "wet soil", "soft rock", "hard rock", and "ice".
    K1 : FloatLike, optional
        Variable used in crater scaling (see Richardson 2009)
    mu : FloatLike, optional
        Variable used in crater scaling (see Richardson 2009)
    Ybar : FloatLike, optional
        The strength of the target material, (Pa)
    density : FloatLike, optional
        Volumentric density of target material, (kg/m³)
    monte_carlo_scaling : bool, default=True
        If True, the scaling laws will be applied using monte carlo methods to account for the uncertainty in the scaling laws. If False, the scaling laws will be applied deterministically.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Notes
    -----
    - The `target` parameter is required and must be an instance of the `Target` class.
    - The `material` parameter is optional. If not provided, it will be retrieved from `target`. Setting it explicitly will override the value in `target`.
    - The `K1`, `mu`, `Ybar`, and `density` parameters are optional. If not provided, they will be retrieved from the material catalogue based on the `material`. Setting them explicitly will override the values in the catalogue.
    - The built-in material property values are from Holsapple (1993) and Kraus et al. (2011).
    - Complex crater scaling parameters are a synthesis of Pike (1980), Croft (1985), and Schenk et al. (2004), with updated simple-to-complex transition diameter values from Schenk et al. (2021).

    References
    ----------
    - Richardson, J.E., 2009. Cratering saturation and equilibrium: A new model looks at an old problem. Icarus 204, 697-715. https://doi.org/10.1016/j.icarus.2009.07.029
    - Holsapple, K.A., 1993. The scaling of impact processes in planetary sciences 21, 333-373. https://doi.org/10.1146/annurev.ea.21.050193.002001
    - Kraus, R.G., Senft, L.E., Stewart, S.T., 2011. Impacts onto H2O ice: Scaling laws for melting, vaporization, excavation, and final crater size. Icarus 214, 724-738. https://doi.org/10.1016/j.icarus.2011.05.016
    - Pike, R.J., 1980. Control of crater morphology by gravity and target type - Mars, earth, moon. In: Lunar and Planetary Science Conference 11, 2159-2189.
    - Croft, S.K., 1985. The scaling of complex craters. Proceedings of the Fifteenth Lunar and Planetary Science Conference, Part 2 Journal of Geophysical Research 90, Supplement, C828-C842.
    - Schenk, P.M., Chapman, C.R., Zahnle, K., Moore, J.M., 2004. Ages and interiors: the cratering record of the Galilean satellites, Cambridge University Press. Cambridge University Press, Cambridge, UK.
    - Schenk, P., Castillo-Rogez, J., Otto, K.A., Marchi, S., O'Brien, D., Bland, M., Hughson, K., Schmidt, B., Scully, J., Buczkowski, D., Krohn, K., Hoogenboom, T., Kramer, G., Bray, V., Neesemann, A., Hiesinger, H., Platz, T., De Sanctis, M.C., Schroeder, S., Le Corre, L., McFadden, L., Sykes, M., Raymond, C., Russell, C.T., 2021. Compositional control on impact crater formation on mid-sized planetary bodies: Dawn at Ceres and Vesta, Cassini at Saturn. Icarus 359, 114343. https://doi.org/10.1016/j.icarus.2021.114343

    """

    def __init__(
        self,
        target: Target | str | None = None,
        projectile: Projectile | str | None = None,
        material: str | None = None,
        K1: FloatLike | None = None,
        mu: FloatLike | None = None,
        Ybar: FloatLike | None = None,
        density: FloatLike | None = None,
        monte_carlo_scaling: bool = True,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs,
    ):
        super().__init__(
            target=target,
            projectile=projectile,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )

        object.__setattr__(self, "_K1", None)
        object.__setattr__(self, "_mu", None)
        object.__setattr__(self, "_Ybar", None)
        object.__setattr__(self, "_transition_diameter", None)
        object.__setattr__(self, "_transition_nominal", None)
        object.__setattr__(self, "_complex_enlargement_factor", None)
        object.__setattr__(self, "_simple_enlargement_factor", None)
        object.__setattr__(self, "_final_exp", None)
        object.__setattr__(self, "_material_catalogue", None)
        object.__setattr__(self, "_montecarlo_scaling", monte_carlo_scaling)

        if material is None:
            self.material = self.target.material
        else:
            self.material = material

        _set_properties(
            self,
            material=self.material,
            K1=K1,
            mu=mu,
            Ybar=Ybar,
            catalogue=self.material_catalogue,
            **kwargs,
        )
        if density is not None:
            self.target.density = density
        elif self.target.density is None:
            self.target.density = self.material_catalogue[self.material]["density"]
        if self.projectile.density is None:
            self.projectile.density = self.target.density

        arg_check = sum(x is None for x in [self.target.density, self.K1, self.mu, self.Ybar])
        if arg_check > 0:
            raise ValueError("Scaling model is missing required parameters. Please check the material name and target properties.")
        # Initialize transition factors
        self.recompute()
        return

    def __str__(self) -> str:
        base = super().__str__()
        ybar = format_large_units(self.Ybar, quantity="pressure")
        dt = format_large_units(self.transition_nominal, quantity="length")
        return (
            f"{base}\n"
            f"Material: {self.material}\n"
            f"K1: {self.K1:.3f}\n"
            f"mu: {self.mu:.3f}\n"
            f"Ybar: {ybar}\n"
            f"Target density: {self.target.density:.0f} kg/m³\n"
            f"Projectile density: {self.projectile.density:.0f} kg/m³\n"
            f"Nominal simple-complex transition diameter: {dt}\n"
            f"Monte Carlo Scaling: {self._montecarlo_scaling}"
        )

    def _get_morphology_type(self, diameter: FloatLike | None = None, **kwargs: Any) -> str:
        """
        Computes and the morphology type of a crater and returns a string corresponding to its type.

        Parameters
        ----------
        diameter : float
            The diameter of the crater to compute.

        Returns
        -------
        str
            The type of crater "simple", "complex", or "transitional"
        """
        if not isinstance(diameter, FloatLike) or diameter <= 0 or not np.isfinite(diameter):
            raise ValueError("diameter must be a positive finite number")

        # Use the 1/2x to 2x the nominal value of the simple->complex transition diameter to get the range of the "transitional" morphology type. This is supported by: Schenk et al. (2004) and Pike (1980) in particular
        transition_range = (0.5 * self.transition_nominal, 2 * self.transition_nominal)

        if diameter < transition_range[0]:
            morphology_type = "simple"
        elif diameter > transition_range[1]:
            morphology_type = "complex"
        else:
            if self._montecarlo_scaling:
                # We'll uses the distance from the nominal transition diameter to set a probability of being either simple, complex, or transitional.
                if diameter < self.transition_nominal:
                    p = (self.transition_nominal - diameter) / (self.transition_nominal - transition_range[0])
                    categories = ["simple", "transitional"]
                    prob = [p, 1.0 - p]
                    morphology_type = self.rng.choice(categories, p=prob).item()
                else:
                    p = (diameter - self.transition_nominal) / (transition_range[1] - self.transition_nominal)
                    categories = ["complex", "transitional"]
                    prob = [p, 1.0 - p]
                    morphology_type = self.rng.choice(categories, p=prob).item()
            else:
                morphology_type = "transitional"

        return morphology_type

    def final_to_transient(
        self,
        final_diameter: FloatLike | None = None,
        morphology_type: str | None = None,
        **kwargs,
    ) -> float:
        """
        Computes the transient diameter of a crater based on its final diameter and morphology type.

        This method first ensures that the morphology type of the crater is computed. It then calculates
        the transient crater diameter based on the final diameter using scaling factors for simple or complex
        crater morphologies.

        Parameters
        ----------
        final_diameter : float-like
            The final crater diameter in meters for which to compute the transient diameter.
        morphology_type : str, optional
            The morphology type of the crater ("simple", "complex", "transitional")

        Returns
        -------
        float
            Returns the crater transient diameter in meters
        """
        if not isinstance(final_diameter, FloatLike) or final_diameter <= 0 or not np.isfinite(final_diameter):
            raise ValueError("final_diameter must be a positive finite number")
        if not morphology_type:
            morphology_type = self._get_morphology_type(final_diameter)

        if morphology_type == "simple":
            transient_diameter = self._f2t_simple(final_diameter)
        else:
            transient_diameter = self._f2t_complex(final_diameter)

        transient_diameter = float(transient_diameter)
        return transient_diameter, morphology_type

    def transient_to_final(self, transient_diameter: FloatLike, **kwargs: Any) -> tuple[float, str]:
        """
        Computes the final diameter of a crater based on its transient diameter and morphology type.

        This method first ensures that the morphology type of the crater is computed. It then calculates
        the final crater diameter based on the transient diameter using scaling factors for simple or complex
        crater morphologies. This is a bit more complicated than the final->transient calculation  because In
        the transition region, a particular transient crater diameter could be associate with simple, complex,
        or transitional crater morphologies. Therefore we need to monte carlo our way into a solution to avoid
        biasing in favor of one or another in the transient->final computation

        Parameters
        ----------
        transient_diameter : FloatLike
            The transient diameter in meters of the crater to convert to final.

        Returns
        -------
        float
            The final crater diameter
        str
            The morphology type of the crater
        """
        # validate that transient_diameter is number and that it is positive and finite
        if not isinstance(transient_diameter, FloatLike) or transient_diameter <= 0 or not np.isfinite(transient_diameter):
            raise ValueError("transient_diameter must be a positive finite number")

        # Invert the final -> transient functions for  each crater type
        final_diameter_simple = transient_diameter * self.simple_enlargement_factor

        def root_func(final_diameter, Dt, scaling):
            return scaling._f2t_complex(final_diameter) - Dt

        sol = root_scalar(
            lambda x, *args: root_func(x, *args),
            bracket=(0.1 * final_diameter_simple, 10 * final_diameter_simple),
            args=(transient_diameter, self),
        )
        final_diameter_complex = sol.root

        # Evaluate the potential morphology that this transient crater could be consistent with. If both potential diameter values are unambigusously simple or complex, go with that.
        # If there is disagreement, then we'll draw the answer from a hat and just check to make sure that final_diameter > transient_diameter
        morphology_options = [
            self._get_morphology_type(final_diameter_simple),
            self._get_morphology_type(final_diameter_complex),
        ]

        if len(set(morphology_options)) == 1:  # We have agreement!
            morphology_type = morphology_options[0]
            if morphology_type == "simple":
                final_diameter = final_diameter_simple
            else:
                final_diameter = final_diameter_complex  # this includes transitional types as well
        else:
            if "simple" in morphology_options:  # The disagreement is between simple/complex or simple/transitional
                if morphology_options[0] == "simple":
                    sind = 0
                    cind = 1
                else:
                    sind = 1
                    cind = 0

                if self._montecarlo_scaling:
                    # Randomly draw a morphology based on weighting by whichever option is closest to the transition
                    p = self.rng.random()
                else:
                    p = 0.5
                is_simple = p < abs(final_diameter_complex - self.transition_diameter) / abs(
                    final_diameter_simple - final_diameter_complex
                )
                if is_simple:
                    final_diameter = final_diameter_simple
                    morphology_type = morphology_options[sind]
                else:
                    final_diameter = final_diameter_complex
                    morphology_type = morphology_options[cind]
            else:
                final_diameter = final_diameter_complex
                if self._montecarlo_scaling:
                    morphology_type = self.rng.choice(morphology_options)
                else:
                    morphology_type = "transitional"

        final_diameter = float(final_diameter)
        morphology_type = morphology_type
        return final_diameter, morphology_type

    def projectile_to_transient(self, projectile_diameter: FloatLike, **kwargs: Any) -> float:
        """
        Calculate the transient diameter of a crater based on the properties of the projectile and target.

        Returns
        -------
        float
            The calculated transient diameter of the crater resulting from the impact.
        """
        if not isinstance(projectile_diameter, FloatLike) or projectile_diameter <= 0 or not np.isfinite(projectile_diameter):
            raise ValueError("projectile_diameter must be a positive finite number")

        # Compute some auxiliary quantites
        projectile_radius = projectile_diameter / 2
        projectile_mass = (4.0 / 3.0) * math.pi * (projectile_radius**3) * self.projectile.density

        c1 = 1.0 + 0.5 * self.mu
        c2 = (-3 * self.mu) / (2.0 + self.mu)

        # Find dimensionless quantities
        pitwo = (self.target.gravity * projectile_radius) / (self.projectile.vertical_velocity**2)
        pithree = self.Ybar / (self.target.density * (self.projectile.vertical_velocity**2))
        pifour = self.target.density / self.projectile.density
        pivol = self.K1 * ((pitwo * (pifour ** (-1.0 / 3.0))) + (pithree**c1)) ** c2
        pivolg = self.K1 * (pitwo * (pifour ** (-1.0 / 3.0))) ** c2

        # find transient crater volume and radii (depth = 1/3 diameter)
        cvol = pivol * (projectile_mass / self.target.density)
        cvolg = pivolg * (projectile_mass / self.target.density)
        transient_radius = (3 * cvol / math.pi) ** (1.0 / 3.0)
        # TODO: transient_radius_gravscale = (3 * cvolg / math.pi)**(1.0/3.0)

        transient_diameter = transient_radius * 2

        if transient_diameter < projectile_diameter:
            transient_diameter = projectile_diameter

        return transient_diameter

    def transient_to_projectile(self, transient_diameter: FloatLike, **kwargs: Any) -> float:
        """
        Estimate the characteristics of the projectile that could have created a given crater.

        This method approximates the properties of a hypothetical projectile based on the characteristics
        of a known crater.

        Parameters
        ----------
        transient_diameter : float
            The diameter of the crater in meters.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        Crater
            The computed projectile for the crater.
        """
        if not isinstance(transient_diameter, FloatLike) or transient_diameter <= 0 or not np.isfinite(transient_diameter):
            raise ValueError("transient_diameter must be a positive finite number")

        def root_func(projectile_diameter) -> float:
            value = self.projectile_to_transient(projectile_diameter)
            return value - transient_diameter

        sol = root_scalar(
            lambda x, *args: root_func(x, *args),
            bracket=(1e-5 * transient_diameter, 1.2 * transient_diameter),
        )

        return sol.root

    @property
    def material_catalogue(self):
        """
        The material catalogue used to look up material properties. Material property values are from Holsapple (1993) and Kraus et al. (2011).

        Returns
        -------
        dict
            The material catalogue.

        References
        ----------
        - Holsapple, K.A., 1993. The scaling of impact processes in planetary sciences 21, 333-373. https://doi.org/10.1146/annurev.ea.21.050193.002001
        - Kraus, R.G., Senft, L.E., Stewart, S.T., 2011. Impacts onto H2O ice: Scaling laws for melting, vaporization, excavation, and final crater size. Icarus 214, 724-738. https://doi.org/10.1016/j.icarus.2011.05.016
        """

        def _create_material_catalogue():
            # Define some built-in catalogue values for known solar system materials of interest
            # Define some default crater scaling relationship terms (see Richardson 2009, Table 1, and Kraus et al. 2011 for Ice)
            material_properties = ["name", "K1", "mu", "Ybar", "density"]
            material_values = [
                ("Water", 2.30, 0.55, 0.0, 1000.0),
                ("Sand", 0.24, 0.41, 0.0, 1750.0),
                ("Dry Soil", 0.24, 0.41, 0.18e6, 1500.0),
                ("Wet Soil", 0.20, 0.55, 1.14e6, 2000.0),
                ("Soft Rock", 0.20, 0.55, 7.60e6, 2250.0),
                ("Hard Rock", 0.20, 0.55, 18.0e6, 2500.0),
                ("Ice", 15.625, 0.48, 0.0, 900.0),
            ]
            return _create_catalogue(material_properties, material_values)

        if self._material_catalogue is None:
            self._material_catalogue = _create_material_catalogue()
        return self._material_catalogue

    def recompute(self):
        """
        Computes and sets the internal attributes for transition factors between simple and complex craters.
        """
        # These terms are used to compute the ratio of the transient crater to simple crater size
        simple_enlargement_mean = 0.84  # See Melosh (1989) pg. 129 just following eq. 8.2.1
        simple_enlargement_std = 0.04  # Just using Pike (1980) fig. 9 the crater depth varies by about the same amount on either side of the transition so this is a reasonable assumption

        # These terms are used in the exponent in the final rim radius/ simple crater radius vs  final radius / transition radius relationship
        # See Holsapple (1993) eq. 28
        final_exp_mean = 0.079
        final_exp_std = 0.0001  # We add noise because this is nature and nature messy
        complex_enlargement_factor = 1.02

        # These terms are used to compute the transition diameter as a function of gravity. They are based on fits to the plot given in Fig. 7 of Schenk et al. (2021)
        if self.target.transition_scale_type == "silicate":
            simple_complex_A = -0.5484575694575697
            simple_complex_B = 4.201369948094472
            simple_complex_sigma = 0.0699898822890725
        elif self.target.transition_scale_type == "ice":
            simple_complex_A = -0.7066985617230864
            simple_complex_B = 3.52268633631802
            simple_complex_sigma = 0.10337890091526512

        def _sample_transition_diameter(g, A, B, sigma):
            mu = A * np.log10(g) + B
            return 10 ** self.rng.normal(mu, sigma)

        # The nominal value will be used for determining the range of the "transitional" morphology type
        self._transition_nominal = 10 ** (simple_complex_A * np.log10(self.target.gravity) + simple_complex_B)

        # Draw from a truncated normal distribution for each component of the model
        if self._montecarlo_scaling:
            simple_enlargement_factor = 1.0 / mc.bounded_norm(simple_enlargement_mean, simple_enlargement_std, rng=self.rng)[0]
            final_exp = mc.bounded_norm(final_exp_mean, final_exp_std, rng=self.rng)[0]
            transition_diameter = _sample_transition_diameter(
                self.target.gravity,
                simple_complex_A,
                simple_complex_B,
                simple_complex_sigma,
            )
        else:
            simple_enlargement_factor = 1.0 / simple_enlargement_mean
            final_exp = final_exp_mean
            transition_diameter = float(self.transition_nominal)
        self._transition_diameter = float(transition_diameter)
        self._simple_enlargement_factor = float(simple_enlargement_factor)
        self._complex_enlargement_factor = float(complex_enlargement_factor)
        self._final_exp = float(final_exp)
        return

    def _f2t_simple(self, Df):
        return Df / self.simple_enlargement_factor

    def _f2t_complex(self, Df):
        return (
            Df
            / (self.simple_enlargement_factor * self.complex_enlargement_factor)
            * (Df / self.transition_diameter) ** -self.final_exp
        )

    @property
    def transition_diameter(self) -> float:
        """The transition diameter between simple and complex craters in m."""
        return self._transition_diameter

    @property
    def transition_nominal(self) -> float:
        """The nominal transition diameter between simple and complex craters in m."""
        return self._transition_nominal

    @property
    def simple_enlargement_factor(self) -> float:
        """The enlargement factor for simple craters."""
        return self._simple_enlargement_factor

    @property
    def complex_enlargement_factor(self) -> float:
        """The enlargement factor for complex craters."""
        return self._complex_enlargement_factor

    @property
    def final_exp(self) -> float:
        """The exponent used in the final rim radius to simple crater radius relationship."""
        return self._final_exp

    @parameter
    def K1(self):
        """The K₁ term in the crater scaling relationship."""
        return self._K1

    @K1.setter
    def K1(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("K1 must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("K1 must be a positive number")
        self._K1 = float(value)

    @parameter
    def mu(self):
        """The μ term in the crater scaling relationship."""
        return self._mu

    @mu.setter
    def mu(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("mu must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("mu must be a positive number")
        self._mu = float(value)

    @parameter
    def Ybar(self):
        """The Ȳ term in the crater scaling relationship, which is the strength of the material in Pa."""
        return self._Ybar

    @Ybar.setter
    def Ybar(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("Ybar must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("Ybar must be a positive number")
        self._Ybar = float(value)

    @property
    def catalogue_key(self):
        """The key used to identify the property used as the key in a catalogue."""
        return "material"

    @parameter
    def material(self):
        """The name of the material composition of the target body."""
        return self._material

    @material.setter
    def material(self, value):
        if value is None:
            return
        if not isinstance(value, str):
            raise TypeError("name must be a string or None")
        self._material = value.title()
        if self._material in self.material_catalogue:
            self.K1 = self.material_catalogue[self._material]["K1"]
            self.mu = self.material_catalogue[self._material]["mu"]
            self.Ybar = self.material_catalogue[self._material]["Ybar"]
            self.target.density = self.material_catalogue[self._material]["density"]

    @parameter
    def monte_carlo_scaling(self):
        """If True, the scaling laws will be applied using monte carlo methods to account for the uncertainty in the scaling laws. If False, the scaling laws will be applied deterministically."""
        return self._montecarlo_scaling

    @monte_carlo_scaling.setter
    def monte_carlo_scaling(self, value):
        if not isinstance(value, bool):
            raise TypeError("monte_carlo_scaling must be a boolean")
        self._montecarlo_scaling = value



from collections.abc import Sequence
from typing import Any

import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike, NDArray

from cratermaker.components.counting import R_to_CSFD
from cratermaker.components.production import Production
from cratermaker.constants import FloatLike
from cratermaker.utils.general_utils import format_large_units, parameter


@Production.register("neukum")
class NeukumProduction(Production):
    """
    An operations class for computing the the Neukum production function for the Moon and Mars.

    Notes
    -----
    The CSFD is computed using the model of Ivanov, Neukum, and Hartmann (2001) for the Moon and Mars, with
    minor changes. Notably, there is a typo in the chronology function (Eq. 5) of the original paper. The linear term in the paper
    is given as 8.38e-4. The value should be 10^(a0), and therefore the number given in the paper is based on the "Old"
    coefficients from Neukum (1983). The correct value is 10^(-3.0876) = 8.17e-4. We compute the value from the coefficients
    in our implementation of the chronology function.

    The Mars production function is based on Ivanov (2001). The projectile production function is from Ivanov, Neukum, and Wagner (2001).

    References
    ----------
    - Neukum, G., Ivanov, B.A., Hartmann, W.K., 2001. Cratering Records in the Inner Solar System in Relation to
      the Lunar Reference System. *Space Science Reviews*, 96, 55-86. https://doi.org/10.1023/A:1011989004263
    - Ivanov, B.A., 2001. Mars/Moon Cratering Rate Ratio Estimates. *Space Science Reviews*, 96, 87-104. https://doi.org/10.1023/A:1011941121102
    - Ivanov, B.A., Neukum, G., Wagner, R., 2001. Size-Frequency Distributions of Planetary Impact Craters
      and Asteroids. In *Collisional Processes in the Solar System*, Springer Netherlands, Dordrecht, pp. 1-34.
      https://doi.org/10.1007/978-94-010-0712-2_1
    - Neukum, G., 1983. Meteorite bombardment and planetary surfaces dating. Habilitation Dissertation for Faculty Membership, University of Munich.

    """

    def __init__(
        self,
        version: str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        version : {"Moon", "Mars", "Projectile"}, optional
            The specific model to use for the production function. "Moon" and "Mars" are both crater production functions, and
            "Projectile" is a projectile function. Default is "Moon".
        rng : numpy.random.Generator | None
            |rng|
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            |rng_seed|
        rng_state : dict, optional
            |rng_state|
        **kwargs : Any
            |kwargs|
        """
        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_Cexp", None)
        object.__setattr__(self, "_Clin", None)
        object.__setattr__(self, "_tau", None)

        self.version = version or "Moon"

    def __str__(self) -> str:
        base = super().__str__()
        timelo = format_large_units(self.valid_time[0], quantity="time")
        timehi = format_large_units(self.valid_time[1], quantity="time")
        dlo = format_large_units(self.sfd_range[0], quantity="length")
        dhi = format_large_units(self.sfd_range[1], quantity="length")
        return f"{base}\nVersion: {self.version}\nValid Time Range: {timelo} - {timehi}\nValid Diameter Range: {dlo} - {dhi}"

    def function(
        self,
        diameter: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        time_start: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        time_end: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> FloatLike | ArrayLike:
        """
        Return the cumulative size-frequency distribution of craters over a given age range and crater diameter.

        Parameters
        ----------
        diameter : FloatLike or numpy array
            Crater diameter(s) in units of meters to compute corresponding cumulative number density value.
        time_start : FloatLike or ArrayLike, default=1.0
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        time_end : FloatLike or ArrayLike, optional
            The ending age in units of My relative to the present, which is used to compute the cumulative SFD. The default is 0 (present day).
        validate_inputs: bool, default=True
            If True, the function will check that the validity of age, time_end, and diameter arguments. If False, no check is performed, and the arguments are assumed to be valid. This can be used to speed up the function, particularly if it is called as part of a solver or optimization routine where the inputs are already known to be valid.

        Returns
        -------
        FloatLike or numpy array of FloatLike
            The cumulative number of craters per square meter greater than the input diameter that would be expected to form on a
            surface over the given age range.
        """
        if "age" in kwargs:
            time_start = kwargs.pop("age")
        if validate_inputs:
            time_start, time_end = self._validate_age(time_start, time_end)
            diameter, _ = self._validate_csfd(diameter=diameter)

        diameter_array = self.csfd(diameter)
        age_difference = self.chronology(time_start=time_start, time_end=time_end, validate_inputs=validate_inputs)

        if diameter_array.ndim > 0 and age_difference.ndim > 0:
            return diameter_array[:, None] * age_difference
        else:
            return diameter_array * age_difference

    def chronology(
        self,
        time_start: ArrayLike = (1000.0),
        time_end: ArrayLike | None = None,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> NDArray[np.float64]:
        """
        Returns the relative number of craters produced over a given age range.

        This implements the chronology function given in
        Eq. 5 of Ivanov, Neukum, and Hartmann (2001) SSR v. 96 pp. 55-86, but takes in the age argument in the Cratermaker unit
        system of My instead of Gy.

        Parameters
        ----------
        age : ArrayLike, default=1000.0
            Age in the past relative to the present day to compute cumulative SFD in units of My.
        time_end : ArrayLike, default=0.0
            The ending age in the past relative to the present day to compute cumulative SFD in units of My. The default is 0 (present day).
        validate_inputs : bool, default=True
            If True, the function will check that the age is within the valid range of ages for this production function. If False, no check is performed.

        Returns
        -------
        NDArray
            The number of craters relative to the amount produced in the last 1 My between age and ange_end.

        """
        if "age" in kwargs:
            time_start = kwargs.pop("age")
        if validate_inputs:
            time_start, time_end = self._validate_age(time_start, time_end)

        def _N1km(
            age: Sequence[FloatLike] | ArrayLike,
        ) -> FloatLike | ArrayLike:
            """
            Return the cumulative number of 1 km craters as a function of age in Gy.

            This is a direct implementation of Eq. 5 in Ivanov, Neukum, and Hartmann (2001) SSR v. 96 pp. 55-86 (with corrected coefficient for the linear term).

            Parameters
            ----------
            age : FloatLike or numpy array
                Time ago in units of My

            Returns
            -------
            FloatLike or numpy array
                The number of craters per square meter greater than 1 km in diameter
            """
            N1 = self.Cexp * (np.exp(age * 1e-3 / self.tau) - 1.0) + self.Clin * age * 1e-3
            return N1 * 1e-6

        if time_end is None:
            return _N1km(age=time_start)
        else:
            return _N1km(age=time_start) - _N1km(age=time_end)

    @property
    def valid_versions(self) -> list[str]:
        """
        The valid versions of the production function.

        Returns
        -------
        list
            The valid versions of the production function.
        """
        return ["Moon", "Mars", "Projectile"]

    @property
    def generator_type(self) -> str:
        """
        Get the generator type of the production function.

        Returns
        -------
        str
            The generator type of the production function.
        """
        return self._generator_type

    @generator_type.setter
    def generator_type(self, value: str) -> None:
        """
        The generator type for this production function is determined by the version.
        """
        raise NotImplementedError("The generator type cannot be changed for this production function.")

    @parameter
    def version(self) -> str:
        """
        Get the version of the production function.

        Returns
        -------
        str
            The version of the production function.
        """
        return self._version

    @version.setter
    def version(self, value: str) -> None:
        """
        Set the version of the production function.

        Parameters
        ----------
        value : str
            The version of the production function.
        """
        if value.title() not in self.valid_versions:
            raise ValueError(f"Invalid version '{value}'. Valid options are {self.valid_versions}")
        self._version = value.title()

        if self._version == "Projectile":
            self._generator_type = "projectile"
        else:
            self._generator_type = "crater"

    @property
    def sfd_tables(self) -> dict[str, list[float]]:
        """
        Get the table of coefficients for the production function.

        Returns
        -------
        dict
            The table of coefficients for the production function.
        """
        return {
            "Moon": [
                -3.0876,
                -3.557528,
                +0.781027,
                +1.021521,
                -0.156012,
                -0.444058,
                +0.019977,
                +0.086850,
                -0.005874,
                -0.006809,
                +8.25e-4,
                +5.54e-5,
            ],
            "Mars": [
                -3.384,
                -3.197,
                +1.257,
                +0.7915,
                -0.4861,
                -0.3630,
                +0.1016,
                +6.756e-2,
                -1.181e-2,
                -4.753e-3,
                +6.233e-4,
                +5.805e-5,
            ],
            "Projectile": [
                0,
                +1.375458,
                +1.272521e-1,
                -1.282166,
                -3.074558e-1,
                +4.149280e-1,
                +1.910668e-1,
                -4.260980e-2,
                -3.976305e-2,
                -3.180179e-3,
                +2.799369e-3,
                +6.892223e-4,
                +2.614385e-6,
                -1.416178e-5,
                -1.191124e-6,
            ],
        }

    @property
    def sfd_coef(self) -> list[float]:
        """
        Get the table of coefficients for the production function.

        Returns
        -------
        list
            The table of SFD coefficients for the production function.
        """
        return self.sfd_tables[self.version]

    @property
    def Clin(self) -> float:
        """
        Get the linear coefficient for the production function.

        Returns
        -------
        float
            The linear coefficient for the production function.
        """
        if self._Clin is None:
            self._Clin = 10 ** (self.sfd_coef[0])
        return self._Clin

    @property
    def Cexp(self) -> float:
        """
        Get the exponential coefficient for the production function.

        Returns
        -------
        float
            The exponential coefficient for the production function.
        """
        if self._Cexp is None:
            Cexp_moon = 5.44e-14
            if self.version == "Moon":
                self._Cexp = Cexp_moon
            else:
                Clin_moon = 10 ** (self.sfd_tables["Moon"][0])
                self._Cexp = Cexp_moon * self.Clin / Clin_moon
        return self._Cexp

    @property
    def tau(self) -> float:
        """
        Get the time constant for the production function.

        Returns
        -------
        float
            The time constant for the production function.
        """
        if self._tau is None:
            self._tau = 1.0 / 6.93
        return self._tau

    @property
    def sfd_range(self) -> tuple[float, float]:
        """
        The range of diameters over which the SFD is valid. The range is given in m.

        Returns
        -------
        tuple
            The lower and upper bounds of the SFD range in m.
        """
        sfd_range = {
            "Moon": (10.0, 1000.0e3),
            "Mars": (15.0, 362.0e3),
            "Projectile": (
                0.1,
                200.0e3,
            ),  # Estimated based on Fig. 16 of Ivanov et al. (2001)
        }

        return sfd_range[self.version]

    @property
    def valid_time(self) -> tuple[float, float]:
        """
        The range of ages over which the production function is valid. The range is given in My.

        Returns
        -------
        tuple
            The lower and upper bounds of the valid time range in My.
        """
        return (0, 4500)

    def csfd(
        self,
        diameter: FloatLike | ArrayLike,
    ) -> FloatLike | ArrayLike:
        """
        Return the cumulative size frequency distribution of craters at a given age relative to age = 1 My ago per m².

        Parameters
        ----------
        diameter : FloatLike or ArrayLike
            units of meter

        Returns
        -------
        FloatLike or numpy array
           Cumulative number density of per square meter greater than the input diameter.
        """
        # Convert m to km for internal functions
        Dkm_lo = self.sfd_range[0] * 1e-3
        Dkm_hi = self.sfd_range[1] * 1e-3

        def _extrapolate_sfd(side: str = "lo") -> FloatLike | ArrayLike:
            """
            Return the exponent, p, and and proportionality constant, A, for  the extrapolated CSFD in the form N(D) = A * D**-p.

            Parameters
            ----------
            side : str
                The side of the range to extrapolate. Valid values are "lo" and "hi"

            Returns
            -------
            A, p
            """
            if side == "lo":
                Dkm = Dkm_lo
            elif side == "hi":
                Dkm = Dkm_hi
            else:
                raise ValueError("side must be 'lo' or 'hi'")
            p = _dNdD(Dkm)
            A = _CSFD(Dkm)
            return A, p

        def _dNdD(
            Dkm: FloatLike | Sequence[FloatLike] | ArrayLike,
        ) -> FloatLike | ArrayLike:
            """
            Return the derivative of the cumulative size-frequency distribution as a function of diameter.

            For diameter values outside the range of the NPF, the derivative is extrapolated using a power law.

            Parameters
            ----------
            Dkm : FloatLike or numpy array
                diameter in units of km

            Returns
            -------
            FloatLike or numpy array
                The differential number of craters (dN/dD) per square kilometer greater than Dkm in diameter at age = 1 Gy ago.
            """

            def _dNdD_scalar(Dkm):
                dcoef = self.sfd_coef[1:]
                if Dkm < Dkm_lo:
                    A, p = _extrapolate_sfd(side="lo")
                    return A * (p / Dkm) * (Dkm / Dkm_lo) ** p
                elif Dkm > Dkm_hi:
                    A, p = _extrapolate_sfd(side="hi")
                    return A * (p / Dkm) * (Dkm / Dkm_hi) ** p
                else:
                    return sum(co * np.log10(Dkm) ** i for i, co in enumerate(dcoef))

            return _dNdD_scalar(Dkm) if np.isscalar(Dkm) else np.vectorize(_dNdD_scalar)(Dkm)

        def _CSFD(
            Dkm: FloatLike | Sequence[FloatLike] | ArrayLike,
        ) -> FloatLike | ArrayLike:
            """
            Return the cumulative size-frequency distribution at the reference age of 1 Gy ago.

            For diameter values outside the range of the NPF, the CSFD is extrapolated using a power law.

            Parameters
            ----------
            Dkm : FloatLike or numpy array
                diameter in units of km

            Returns
            -------
            FloatLike or numpy array
                The number of craters per square kilometer greater than Dkm in diameter at age=1 Gy ago.
            """

            def _CSFD_scalar(Dkm):
                if Dkm < Dkm_lo:
                    A, p = _extrapolate_sfd(side="lo")
                    return A * (Dkm / Dkm_lo) ** p
                elif Dkm > Dkm_hi:
                    A, p = _extrapolate_sfd(side="hi")
                    p -= 2.0  # Steepen the upper branch of the SFD to prevent anomolously large craters from forming
                    return A * (Dkm / Dkm_hi) ** p
                else:
                    logCSFD = sum(co * np.log10(Dkm) ** i for i, co in enumerate(self.sfd_coef))
                    return 10**logCSFD

            return _CSFD_scalar(Dkm) if np.isscalar(Dkm) else np.vectorize(_CSFD_scalar)(Dkm)

        if np.any(diameter < 0.0):
            raise ValueError("diameter must be greater than or equal to 0.0")

        Dkm = diameter * 1e-3  # Convert m to km for internal functions

        if self.version == "Projectile":
            Ncumulative = (
                R_to_CSFD(R=_CSFD, D=Dkm) * 2.94e-5
            )  # This is a multiplication factor that gets the projectile CSFD to approximately match the lunar crater CSFD
        else:
            Ncumulative = _CSFD(Dkm)

        return Ncumulative * 1e3  # convert from Gy^-1 km^-2 to My^-1 m^-2



from __future__ import annotations

from pathlib import Path
from typing import Any
from warnings import warn

import numpy as np
import xarray as xr
from numpy.typing import NDArray

from cratermaker.components.morphology import Morphology
from cratermaker.components.scaling import Scaling
from cratermaker.components.surface import Surface
from cratermaker.components.surface.hireslocal import HiResLocalSurface
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike, PairOfFloats
from cratermaker.utils.general_utils import (
    format_large_units,
    parameter,
)

_DEFAULT_N_FACES_LOCAL = 1e6


@Surface.register("datasurface")
class DataSurface(HiResLocalSurface):
    """
    A Surface subclass that generates a local region using DEM data.

    Currently implements only lunar LOLA data from the PDS. Other data sources may be added in the future.

    Parameters
    ----------
    local_radius : FloatLike
        The radius of the local region in meters.
    local_location : PairOfFloats
        The longitude and latitude of the location in degrees.
    superdomain_scale_factor : FloatLike, optional
        A factor defining relative size of the face at the antipode of the local region to the face size inside the local region.
        If not provided, construction of the surface will be deferred until the `set_superdomain` method is called.
        If a negative number is provided, it will be computed based on a provided (or default) scaling and morphology model, and will be set so that smallest craters that can be resolved on faces outside the local region could potentially deposit ejecta at the boundary of the local region.
        Default is -1 (which triggers automatic computation based on scaling and morphology models).
    target : Target, optional
        The target body or name of a known target body for the impact simulation. If none provide, it will be either set to the default,
        or extracted from the scaling model if it is provied
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is True.
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        |simdir|
    pix : FloatLike | None, optional
        The approximate face size inside the local region in meters. This will be used to determine the target resolution of the DEM data to be used. The actual resolution may be different based on the available DEM data. Note that if you provide a list of DEM files using the `dem_file_list` parameter, this value will be ignored. If None, is set, and no file(s) are provided, a default resolution that creates approximately 1e6 faces in the local region will be used.
    local_radius : FloatLike
    dem_file_list : list of str | Path, optional
        A list of one or more DEM files or urls links to files to be used for the local region. If not provided, a set of files will be determined based on the location and radius of the region. It will chose a set of files that fully covers the local region with approximately approximately 1e6 faces.
    superdomain_dem_file :  str | Path, optional
        A global DEM files or urs link to a file to be used for the superdomain region. If not provided, a default global DEM file will be used based on the target body.
    **kwargs : Any

    Returns
    -------
    DataSurface
        An instance of the DataSurface class initialized with faces and elevations drawn from DEM data

    Notes
    -----
    Currently this only uses LOLA DEM data for the Moon.

    """

    def __init__(
        self,
        local_radius: FloatLike,
        local_location: PairOfFloats,
        superdomain_scale_factor: FloatLike | None = -1,
        target: Target | str | None = None,
        reset: bool = True,
        regrid: bool = False,
        simdir: str | Path | None = None,
        pix: FloatLike | None = None,
        dem_file_list: list[str] | list[Path] | None = None,
        superdomain_dem_file: str | Path | None = None,
        **kwargs: Any,
    ):
        try:
            import rasterio  # noqa: F401
        except ImportError as e:
            # IMPORTANT: do not `return` from __init__.
            # Returning leaves a partially-initialized object that will later fail
            # with missing base-class attributes (e.g., _simdir, _output_dir_name).
            raise ImportError(
                "DataSurface requires the optional dependency 'rasterio'. "
                "Install it (and GDAL if needed) to use the 'datasurface' surface type."
            ) from e
        object.__setattr__(self, "_local", None)

        # Temporary storage for DEM data during initialization. This will be cleared after the elevation points are set.
        object.__setattr__(self, "_local_dem_data", None)
        object.__setattr__(self, "_global_dem_data", None)

        # The location of the saved DEM data that can be retrieved later
        object.__setattr__(self, "_dem_output_file", None)
        object.__setattr__(self, "_dem_file_list", None)
        object.__setattr__(self, "_superdomain_dem_file", None)

        super(HiResLocalSurface, self).__init__(target=target, simdir=simdir, **kwargs)
        if dem_file_list is None and self.target.name != "Moon":
            raise ValueError("DataSurface currently only supports the Moon as a target if 'dem_file_list' is not provided.")
        if superdomain_dem_file is None and self.target.name != "Moon":
            raise ValueError("DataSurface currently only supports the Moon as a target if 'superdomain_dem_file' is not provided.")
        self._output_file_pattern += [f"local_{self._output_file_prefix}*.{self._output_file_extension}"]
        self._dem_output_file = f"dem_data.{self._output_file_extension}"

        # Set the attributes directly to avoid triggering checks before the pix value is set
        self._local_radius = local_radius
        self._local_location = local_location
        self._pix = pix
        self.dem_file_list = dem_file_list
        # Temporarily disable ask_overwrite to avoid prompts during initialization
        ask_overwrite = self.ask_overwrite
        self.ask_overwrite = False
        super_kwargs = {
            **kwargs,
            "pix": self.pix,
            "local_radius": local_radius,
            "local_location": local_location,
            "superdomain_scale_factor": superdomain_scale_factor,
            "target": self.target,
            "reset": reset,
            "regrid": regrid,
            "simdir": self.simdir,
            "ask_overwrite": self.ask_overwrite,
        }
        super().__init__(**super_kwargs)

        self._superdomain_dem_file = superdomain_dem_file
        self.ask_overwrite = ask_overwrite
        return

    def _get_lola_dem_file_list(
        self,
        pix: FloatLike,
        lat_range: tuple[float, float],
    ) -> list[str] | str:
        """
        Determine the set of LOLA DEM files needed to cover the local region if none are provided by the user.

        Parameters
        ----------
        pix : FloatLike
            The approximate resolution in meters per pixel to use to select the DEM files.
        lat_range : tuple of float, optional
            The (min_lat, max_lat) in degrees of the local region. If None, it will be computed from the local location and radius.
        """
        target_pds_resolution = np.pi / 180.0 * self.target.radius / pix
        if target_pds_resolution > 10 and (
            lat_range[0] > 60 or lat_range[1] < -60
        ):  # Use polar files high latitude, high resolution regions.
            return self._get_lola_polar_files_from_pds(pix, lat_range=lat_range)
        else:  # Use cylindrical for all other cases
            return self._get_lola_cylindrical_files_from_pds(resolution=target_pds_resolution)

    def _get_location_extents(self):
        """
        Computes the longitude and latitude extents of the local region.

        This is used to determine which DEM files are needed to cover the region.

        Returns
        -------
        lon_min : float
            Minimum longitude in degrees.
        lon_max : float
            Maximum longitude in degrees.
        lat_min : float
            Minimum latitude in degrees.
        lat_max : float
            Maximum latitude in degrees.

        """
        R = self.target.radius
        lon0, lat0 = self.local_location
        alpha_deg = np.degrees(self.local_radius / R)

        lat_min = max(-90.0, lat0 - alpha_deg)
        lat_max = min(90.0, lat0 + alpha_deg)

        # If we hit a pole, longitudes span everything
        if abs(lat0) + alpha_deg >= 90.0:
            lon_min, lon_max = -180.0, 180.0
        else:
            half_lon_span = alpha_deg / np.cos(np.radians(lat0))
            half_lon_span = min(half_lon_span, 180.0)

            lon_min = lon0 - half_lon_span
            lon_max = lon0 + half_lon_span
        return float(lon_min), float(lon_max), float(lat_min), float(lat_max)

    @staticmethod
    def _get_lola_cylindrical_url_from_pds(pds_file_resolution: int, location: PairOfFloats, boundary_offset=(0, 0)) -> Path:
        """
        Retrieve the appropriate LOLA DEM file url for a given location and resolution from the PDS.

        Parameters
        ----------
        pds_file_resolution : int
            DEM resolution in meters per pixel.
        location : PairOfFloats,
            The longitude and latitude of the location in degrees.
        boundary_offset : tuple of int, optional
            Offset to apply to tile index to access neighboring tiles. Default is (0, 0). This is used when the local region crosses one or more tile boundaries.

        Returns
        -------
        url: str
            Link to the DEM file

        Notes
        -----
        This is meant to be used for mid-latitude regions on the Moon between -60 and 60 degrees latitude.
        For resolutions of 128, 256, and 512 pix/deg, this will return the SLDEM2015 datasets.

        """
        lola_cylindrical_url = (
            "https://pds-geosciences.wustl.edu/lro/lro-l-lola-3-rdr-v1/lrolol_1xxx/data/lola_gdr/cylindrical/float_img/"
        )
        sldem_global_url = "https://pds-geosciences.wustl.edu/lro/lro-l-lola-3-rdr-v1/lrolol_1xxx/data/sldem2015/global/float_img/"
        sldem_tile_url = "https://pds-geosciences.wustl.edu/lro/lro-l-lola-3-rdr-v1/lrolol_1xxx/data/sldem2015/tiles/float_img/"

        def _get_pds_file_suffix(pds_file_resolution, location, boundary_offset):
            if location[0] < 0:
                location = (location[0] + 360, location[1])
            if pds_file_resolution == 512:
                dlon = 45
                dlat = 30
            elif pds_file_resolution == 256:
                dlon = 120
                dlat = 60

            latdir = "n" if location[1] + boundary_offset[1] * dlat > 0 else "s"
            latval = np.abs(location[1] / dlat) + boundary_offset[1]
            if latdir == "n":
                latlo = int(np.ceil(latval - 1) * dlat)
                lathi = int(latlo + dlat)
            else:
                lathi = int(np.floor(latval + 1) * dlat)
                latlo = int(lathi - dlat)

            lonval = np.abs(location[0] / dlon) + boundary_offset[0]

            # Pole crossing, so flip the longitudes by 180 degrees
            if lathi > 90 or latlo > 90:
                lathi = 90
                latlo = lathi - dlat
                lonval += 180 / dlon

            lonlo = int(np.floor(lonval) * dlon) % 360
            lonhi = int(lonlo + dlon)

            # Check to make sure the lo and hi values are in the right direction
            if (latdir.upper() == "N" and latlo > lathi) or (latdir.upper() == "S" and latlo < lathi):
                tmp = latlo
                latlo = lathi
                lathi = tmp

            if pds_file_resolution == 512:
                return f"{latlo:02d}{latdir.lower()}_{lathi:02d}{latdir.lower()}_{lonlo:03d}_{lonhi:03d}"
            elif pds_file_resolution == 256:
                return f"{latlo:d}{latdir.lower()}_{lathi:d}{latdir.lower()}_{lonlo:03d}_{lonhi:03d}"

        if pds_file_resolution >= 256:
            url = f"{sldem_tile_url}sldem2015_{pds_file_resolution:d}_{_get_pds_file_suffix(pds_file_resolution, location, boundary_offset)}_float.xml"
        elif pds_file_resolution == 128:
            url = f"{sldem_global_url}sldem2015_{pds_file_resolution:d}_60s_60n_000_360_float.xml"
        else:
            url = f"{lola_cylindrical_url}ldem_{pds_file_resolution:d}_float.xml"

        return url

    def _get_lola_cylindrical_files_from_pds(self, resolution) -> list[str] | str:
        """
        Retrieve the appropriate cylindrically projected LOLA DEM file url or list of urls for a given location and resolution from the PDS.

        This will determine which, if any, boundaries are crossed and will return a list of files to cover the entire local region.

        Parameters
        ----------
        resolution : FloatLike
            Requested resolution in degrees per pixel. The closest available resolution will be used.

        """
        try:
            import rasterio
        except ImportError:
            warn("rasterio is not installed. Cannot use this feature.", stacklevel=2)
            return

        from affine import Affine
        from rasterio.io import MemoryFile
        from rasterio.merge import merge
        from rasterio.vrt import WarpedVRT
        from rasterio.warp import Resampling
        from rasterio.windows import Window, from_bounds

        AVAILABLE_RESOLUTIONS = [4, 16, 64, 128, 256, 512]  #  pix / deg
        diffs = [abs(resolution - res) for res in AVAILABLE_RESOLUTIONS]
        pds_file_resolution = AVAILABLE_RESOLUTIONS[np.argmin(diffs)]

        # First, retrive the file for the centerpoint:
        filelist = [self._get_lola_cylindrical_url_from_pds(pds_file_resolution, self.local_location)]
        if pds_file_resolution < 256:
            return filelist  # These files cover the entire globe, no need to determine if boundaries are crossed

        lon_min, lon_max, lat_min, lat_max = self._get_location_extents()
        combo = [(lon_min, lat_min), (lon_min, lat_max), (lon_max, lat_min), (lon_max, lat_max)]

        for loc in combo:
            f = self._get_lola_cylindrical_url_from_pds(pds_file_resolution, loc)
            if f not in filelist:
                filelist.append(f)

        return filelist

    def _get_lola_polar_files_from_pds(self, resolution, lat_range: tuple[float, float]) -> list[str] | str:
        """
        Retrieve the appropriate polar projected LOLA DEM file url or list of urls for a given location and resolution from the PDS.

        Parameters
        ----------
        resolution : FloatLike
            Requested resolution in meters per pixel. The closest available resolution will be used.
        lat_range : tuple of float
            The (min_lat, max_lat) in degrees of the local region.

        """
        import rasterio

        src_url = "https://pds-geosciences.wustl.edu/lro/lro-l-lola-3-rdr-v1/lrolol_1xxx/data/lola_gdr/polar/float_img/"

        AVAILABLE_RESOLUTIONS = [3000, 1000, 400, 240, 200, 120, 100, 80, 60, 40, 30, 20, 10, 5]
        MIN_LAT = [50, 50, 45, 60, 45, 60, 45, 80, 60, 80, 75, 80, 85, 87.5]

        if lat_range[0] > 0:
            pole = "n"
            min_lat_index = 0  # in the northern hemisphere, the minimum latitude defines coverage
        else:
            pole = "s"
            min_lat_index = 1  # in the southern hemisphere, the maximum latitude defines coverage
        combo = [
            (res, minlat)
            for res, minlat in zip(AVAILABLE_RESOLUTIONS, MIN_LAT, strict=False)
            if abs(lat_range[min_lat_index]) >= minlat
        ]
        if len(combo) == 0:
            raise ValueError("No available LOLA polar DEM files cover the requested latitude range.")

        valid_resolutions, valid_min_lat = zip(*combo, strict=False)

        diffs = [abs(resolution - res) for res in valid_resolutions]
        pds_file_resolution = valid_resolutions[np.argmin(diffs)]
        pds_lat_min = valid_min_lat[np.argmin(diffs)]
        filename = f"ldem_{pds_lat_min:d}{pole}_{pds_file_resolution:d}m"
        url = f"{src_url}{filename}_float.xml"

        self._pix = pds_file_resolution
        return [url]

    def _get_local_dem_data(
        self,
    ):
        """
        Read and extract DEM data from one or more files for a specified region.

        Returns
        -------
        dem : dict or None
            Dictionary with x, y, z coordinates and elevation, or None if boundary case.

        """
        try:
            import rasterio
        except ImportError:
            warn("rasterio is not installed. Cannot use this feature.", stacklevel=2)
            return
        from affine import Affine
        from pyproj import Transformer
        from rasterio.io import MemoryFile
        from rasterio.merge import merge
        from rasterio.vrt import WarpedVRT
        from rasterio.warp import Resampling
        from rasterio.windows import Window, from_bounds

        if self.dem_file_list is None:
            raise ValueError("No DEM files provided to extract data from.")

        _EXPANSION_BUFFER = 1.05
        _NODATA = np.finfo(np.float32).min
        region_radius = self.local_radius
        box_size = 2 * np.sqrt(2.0) * region_radius * _EXPANSION_BUFFER
        half_box_size = box_size / 2
        dst_crs = self.get_crs(
            radius=self.target.radius,
            name=self.target.name,
            location=self.local_location,
        )
        src_list = []
        print("Reading DEM files:")
        try:
            for f in self.dem_file_list:
                print(f"  {f}")
                src_list.append(rasterio.open(f))
        except Exception as e:
            raise RuntimeError(f"Error reading DEM file(s): {e}") from e
        nodata_val = src_list[0].nodata
        if nodata_val is None or np.isnan(nodata_val) or np.abs(nodata_val) < np.abs(_NODATA):
            nodata_val = _NODATA
        target_res = min(s.res[0] for s in src_list)
        dst_width = int(np.ceil(2 * half_box_size / target_res))
        dst_height = dst_width
        dst_transform = Affine(target_res, 0.0, -half_box_size, 0.0, -target_res, half_box_size)
        vrt_list = [
            WarpedVRT(
                src,
                crs=dst_crs,
                transform=dst_transform,
                width=dst_width,
                height=dst_height,
                resampling=Resampling.bilinear,
            )
            for src in src_list
        ]
        # Ensure nodata is respected; if missing, use NaN and float32
        mosaic, transform = merge(
            vrt_list,
            nodata=nodata_val,
            dtype="float32" if np.isnan(nodata_val) else None,
        )

        out_meta = src_list[0].meta.copy()
        out_meta.update(
            {
                "driver": "GTiff",
                "height": mosaic.shape[1],
                "width": mosaic.shape[2],
                "transform": transform,
                "crs": dst_crs,
                "nodata": nodata_val,
            }
        )

        with MemoryFile() as memfile, memfile.open(**out_meta) as src:
            src.units = src_list[0].units
            src.write(mosaic)

            window = Window(0, 0, src.width, src.height)
            if "KILOMETER" in src.units:
                scale_factor = 1000.0
            else:
                scale_factor = 1.0

            # Read the data within the window
            elevation = src.read(1, window=window) * scale_factor

            # Preserve the window grid and affine for later interpolation
            window_transform = src.window_transform(window)

            # Generate row and column indices
            rows, cols = np.indices(elevation.shape)

            # Compute the x and y coordinates of each pixel in the local CRS
            x_coords, y_coords = rasterio.transform.xy(window_transform, rows, cols, offset="center")
            x_coords = np.array(x_coords).flatten()
            y_coords = np.array(y_coords).flatten()
            elevation = elevation.flatten()

            # Handle nodata values
            mask_nodata = (elevation != src.nodata) & ~np.isnan(elevation)
            mean_elevation = np.mean(elevation[mask_nodata])
            elevation[~mask_nodata] = mean_elevation
            elevation = elevation.astype(np.float32)

            transformer_to_geodetic = Transformer.from_crs(dst_crs, self.crs, always_xy=True)

            # Transform to longitude and latitude
            longitudes, latitudes = transformer_to_geodetic.transform(x_coords, y_coords)
            r_vals = self.compute_distances(
                reference_location=self.local_location, locations=list(zip(longitudes, latitudes, strict=False))
            )
            local_dem_data = {
                "elevation": elevation,
                "mask": r_vals <= region_radius + self.pix / 4,
                "latitudes": latitudes,
                "longitudes": longitudes,
            }

        self._local_dem_data = local_dem_data
        return

    def _generate_face_distribution(self, **kwargs: Any) -> NDArray:
        """
        Creates the points that define the mesh centers.

        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.

        """

        def _interior_distribution():
            self._get_local_dem_data()
            mask = self._local_dem_data["mask"]
            longitudes = np.radians(self._local_dem_data["longitudes"][mask])
            latitudes = np.radians(self._local_dem_data["latitudes"][mask])
            x = self.target.radius * np.cos(latitudes) * np.cos(longitudes)
            y = self.target.radius * np.cos(latitudes) * np.sin(longitudes)
            z = self.target.radius * np.sin(latitudes)

            points = np.column_stack([x, y, z]) / self.target.radius

            return points

        def _exterior_distribution(theta):
            r = self.radius
            arc_distance = r * theta
            pix_local = self.superdomain_function(arc_distance)
            if (np.pi * r - arc_distance) < pix_local:
                return [], np.pi

            theta_next = theta + pix_local / r

            sin_theta = np.sin(theta)
            cos_theta = np.cos(theta)

            n_phi = max(1, int(round(2 * np.pi * r * sin_theta / pix_local)))
            phi = np.linspace(0, 2 * np.pi, n_phi, endpoint=False)

            x = sin_theta * np.cos(phi)
            y = sin_theta * np.sin(phi)
            z = np.full_like(phi, cos_theta)

            points = np.column_stack((x, y, z))
            return points.tolist(), theta_next

        print(f"Center of local region: {self.local_location}")
        print(f"Radius of local region: {format_large_units(self.local_radius, quantity='length')}")
        print(f"Local region pixel size: {format_large_units(self.pix, quantity='length')}")

        interior_points = _interior_distribution()
        print(f"Generated {len(interior_points)} points in the local region.")

        theta = (self.local_radius + 0.5 * self.pix) / self.radius
        exterior_points = []
        while theta < np.pi:
            new_points, theta = _exterior_distribution(theta)
            exterior_points.extend(new_points)

        print(f"Generated {len(exterior_points)} points in the superdomain region.")
        exterior_points = self._rotate_point_cloud(exterior_points)

        points = np.vstack([interior_points, exterior_points])
        decimals = max(6, -int(np.floor(np.log10(self.pix / self.radius))))

        points = np.array(points, dtype=np.float64)
        points = np.round(points, decimals=decimals)
        points = np.unique(points, axis=0).T

        return points

    def reset(self, **kwargs: Any) -> None:
        """
        Reset the surface to its initial state.

        Parameters
        ----------
        **kwargs : Any
            |kwargs|

        """
        super().reset(**kwargs)
        if self._local_dem_data is not None:
            self._add_local_dem_elevation()
            self._add_global_dem_elevation()
        elif (self.output_dir / self._dem_output_file).exists():
            with xr.open_dataset(self.output_dir / self._dem_output_file) as ds:
                elevation = np.concatenate([ds.isel(interval=0).face_elevation, ds.isel(interval=0).node_elevation])
                self.update_elevation(elevation)

        return

    def _regrid_if_needed(self, regrid: bool = False, **kwargs: Any) -> bool:
        """
        Check if the existing grid matches the desired parameters determine if regridding is necessary.

        This checks if the dem file exists and matches the current grid.

        Parameters
        ----------
        regrid: bool, optional
            Flag to force regridding even if the grid file exists. Default is False.

        Returns
        -------
        bool
            A boolean indicating whether the grid should be regenerated.
        """
        # DEM data can come from one of two sources. If we are in the middle of building the surface as part of _generate_face_distribution, then the local DEM data should be stored in the _local_dem_data attribute.
        # If this is not set but we already have a grid generated and stored in the uxgrid attribute, then we need to check if a matching DEM file exists on disk. If neither of these sources are avaailable, then we need to regrid.
        if not regrid and self._local_dem_data is None:
            if not (self.output_dir / self._dem_output_file).exists():
                regrid = True
            elif self.uxgrid is not None:
                with xr.open_dataset(self.output_dir / self._dem_output_file) as ds:
                    if (
                        "face_elevation" not in ds
                        or "node_elevation" not in ds
                        or ds.face_elevation.shape[-1] != self.n_face
                        or ds.node_elevation.shape[-1] != self.n_node
                    ):
                        print(f"DEM datafile {self._dem_output_file} does not match current grid. Regridding needed.")
                        regrid = True

        return super()._regrid_if_needed(regrid=regrid, **kwargs)

    def _add_local_dem_elevation(self):
        """
        Sample the preserved DEM grid at the local face and node centers using bilinear interpolation.

        """
        from scipy.interpolate import LinearNDInterpolator

        if self._local_dem_data is None:
            raise ValueError("Local DEM data is not available for interpolation.")
        lonlat = np.c_[self._local_dem_data["longitudes"], self._local_dem_data["latitudes"]]
        lut2 = LinearNDInterpolator(lonlat, self._local_dem_data["elevation"], fill_value=0.0)
        face_elevations = lut2(np.c_[self.local.face_lon, self.local.face_lat])
        node_elevations = lut2(np.c_[self.local.node_lon, self.local.node_lat])
        elevation = np.concatenate([face_elevations, node_elevations])
        self.local.update_elevation(elevation)
        self._local_dem_data = None  # Clear temporary DEM data storage
        return

    def _add_global_dem_elevation(self):
        """
        Sample the preserved DEM grid at the global face and node centers using bilinear interpolation.

        """
        import rasterio
        from pyproj import Transformer
        from scipy.interpolate import LinearNDInterpolator

        ask_overwrite = self.ask_overwrite
        # If we haven't cached global DEM samples yet, build the exterior lon/lat lists
        if self._global_dem_data is None:
            # This will trigger setting the file automatically from the superdomain_scale_factor if not already set.
            self.superdomain_dem_file = self._superdomain_dem_file

        if self._global_dem_data is None:
            self.ask_overwrite = False
            self._global_dem_data = {}

            # Build index arrays for exterior faces/nodes (sorted for reproducibility)
            exterior_face_ind = np.array(sorted(set(range(self.n_face)) - set(self.local.face_indices)), dtype=int)
            exterior_node_ind = np.array(sorted(set(range(self.n_node)) - set(self.local.node_indices)), dtype=int)

            # Cache exterior point coordinates in geodetic lon/lat (degrees)
            self._global_dem_data["face_indices"] = exterior_face_ind
            self._global_dem_data["node_indices"] = exterior_node_ind
            self._global_dem_data["face_longitudes"] = np.asarray(self.face_lon)[exterior_face_ind]
            self._global_dem_data["face_latitudes"] = np.asarray(self.face_lat)[exterior_face_ind]
            self._global_dem_data["node_longitudes"] = np.asarray(self.node_lon)[exterior_node_ind]
            self._global_dem_data["node_latitudes"] = np.asarray(self.node_lat)[exterior_node_ind]

            # Read and sample the global DEM at those exterior points.
            print("Reading global DEM file:")
            print(f"  {self.superdomain_dem_file}")
            with rasterio.open(self.superdomain_dem_file) as dataset:
                if dataset.crs is None:
                    raise ValueError("Global DEM has no CRS; cannot sample elevations.")

                # Transform from geodetic lon/lat (self.crs) into the DEM's CRS
                to_dem = Transformer.from_crs(self.crs, dataset.crs, always_xy=True)

                # Determine unit scaling (match local DEM logic)
                units = getattr(dataset, "units", "") or ""
                if isinstance(units, (list, tuple)):
                    units = units[0] if len(units) else ""
                scale_factor = 1000.0 if "KILOMETER" in str(units).upper() else 1.0

                nodata = dataset.nodata

                def _sample(lons_deg: NDArray, lats_deg: NDArray) -> NDArray:
                    xs, ys = to_dem.transform(lons_deg, lats_deg)
                    # rasterio.sample expects an iterable of (x, y)
                    samples = np.fromiter(
                        (v[0] for v in dataset.sample(zip(xs, ys, strict=False))),
                        dtype=np.float32,
                        count=len(lons_deg),
                    )
                    samples = samples.astype(np.float32) * scale_factor

                    # Replace nodata / NaN with mean of valid samples
                    if nodata is None or np.isnan(nodata):
                        valid = np.isfinite(samples)
                    else:
                        valid = np.isfinite(samples) & (samples != nodata)

                    if np.any(valid):
                        mean_val = float(np.mean(samples[valid]))
                    else:
                        mean_val = 0.0
                    samples[~valid] = mean_val
                    return samples

                face_samples = _sample(self._global_dem_data["face_longitudes"], self._global_dem_data["face_latitudes"])
                node_samples = _sample(self._global_dem_data["node_longitudes"], self._global_dem_data["node_latitudes"])

        # Create full-sized elevation arrays and insert the sampled values at the exterior indices
        face_elevations = np.zeros(self.n_face, dtype=np.float32)
        node_elevations = np.zeros(self.n_node, dtype=np.float32)
        face_elevations[self._global_dem_data["face_indices"]] = face_samples
        node_elevations[self._global_dem_data["node_indices"]] = node_samples

        # Apply elevations to the global (superdomain) surface
        elevation = np.concatenate([face_elevations, node_elevations])
        self.update_elevation(elevation)

        # Save so we can reload later without re-downloading / re-sampling the DEM
        self.save(filename=self._dem_output_file, skip_actions=True)
        self._global_dem_data = None  # Clear temporary DEM data storage
        self.ask_overwrite = ask_overwrite
        return

    @parameter
    def dem_file_list(self) -> int | None:
        """
        The list of files to use for the DEM data in the high resolution local region.

        """
        return self._dem_file_list

    @dem_file_list.setter
    def dem_file_list(self, value: int | None):
        if value is None:
            if self._pix is None:
                # Compute a reasonable default resolution that will contain approximately 1e6 faces based on the local radius
                self._pix = np.sqrt(np.pi * self.local_radius**2 / _DEFAULT_N_FACES_LOCAL)
            value = self._get_lola_dem_file_list(pix=self._pix, lat_range=self._get_location_extents()[2:])
        elif isinstance(value, list):
            if not all(isinstance(f, (str, Path)) for f in value):
                raise ValueError("All items in 'dem_file_list' must be strings or Path objects.")
        elif not isinstance(value, (str, Path)):
            raise TypeError("'dem_file_list' must be a list of strings or Path objects, or None.")

        self._dem_file_list = value

        # Set the pixel size based on the provided files. We take the highest resolution (smallest pixel size) among the files.
        try:
            import rasterio
        except ImportError:
            warn("rasterio is not installed. Cannot use this feature.", stacklevel=2)
            return

        pixvals = []
        for f in self._dem_file_list:
            with rasterio.open(f) as src:
                pixvals.append(src.res[0])
        self._pix = min(pixvals)

        return

    @parameter
    def superdomain_dem_file(self) -> int | None:
        """
        The list of files to use for the DEM data in the superdomain.

        """
        return self._superdomain_dem_file

    @superdomain_dem_file.setter
    def superdomain_dem_file(self, value: int | None):
        if value is None:
            # If the superdomain has not been set yet, we will defer setting this until later
            if self.superdomain_scale_factor is None:
                return
            min_global_pix = np.pi / 180.0 * self.target.radius / 128.0
            sdpix = self.superdomain_scale_factor * self.pix / 10.0
            if sdpix < min_global_pix:
                sdpix = min_global_pix
            self._superdomain_dem_file = self._get_lola_dem_file_list(pix=sdpix, lat_range=(-90, 90))[0]
            return
        if not isinstance(value, (str, Path)):
            raise TypeError("'superdomain_dem_file' must be a strings or Path objects, or None.")

        self._superdomain_dem_file = value

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.

        """
        return super()._hashvars + [self._dem_file_list]



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Sequence
from typing import Any

import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike, NDArray
from scipy.optimize import root_scalar

from cratermaker.components.target import Target
from cratermaker.constants import FloatLike, PairOfFloats
from cratermaker.core.base import ComponentBase, import_components
from cratermaker.utils import montecarlo_utils as mc
from cratermaker.utils.general_utils import parameter


class Production(ComponentBase):
    """
    The base class for computing the production function for craters and projectiles.
    """

    _registry: dict[str, Production] = {}

    def __init__(
        self,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        rng : numpy.random.Generator | None
            |rng|
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            |rng_seed|
        rng_state : dict, optional
            |rng_state|
        **kwargs : Any
            |kwargs|
        """
        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_valid_generator_types", ["crater", "projectile"])

    def __str__(self) -> str:
        base = super().__str__()
        return f"{base}\nGenerator type: {self.generator_type}"

    @classmethod
    def maker(
        cls,
        production: str | Production | None = None,
        target: Target | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ) -> Production:
        """
        Initialize a Production model with the given name or instance.

        Parameters
        ----------
        production : str | Production | None, optional
            The production model to use. This can be either a string or a Production instance.
            If None, the default production model is "neukum" and the version is based on the target (if provided), either Moon, Mars, or Projectile for all other bodies. Default is "Moon"
        target : Target | str | None, optional
            The target body for the impact. Can be a Target object or a string representing the target name.
        rng : numpy.random.Generator | None
            |rng|
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            |rng_seed|
        rng_state : dict, optional
            |rng_state|
        **kwargs : Any
            |kwargs|

        Returns
        -------
        Production
            An instance of the specified production model.

        Raises
        ------
        KeyError
            If the specified production model name is not found in the registry.
        TypeError
            If the specified production model is not a string or a subclass of Production.
        ValueError
            If there is an error initializing the production model.

        """
        set_version = False
        version = None
        if production is None:
            set_version = True
        elif isinstance(production, str):
            production = production.lower()
            if production == "neukum":
                set_version = True
        elif isinstance(production, Production) and production.name.lower() == "neukum":
            set_version = True

        if set_version:
            version = kwargs.pop("version", None)
            target = Target.maker(target, **kwargs)
            if target.name in ["Mercury", "Venus", "Earth", "Moon", "Mars"]:
                production = "neukum"
                if version is None or version != "projectile":
                    if target.name in ["Moon", "Mars"]:
                        version = target.name
                    else:
                        version = "projectile"
            else:
                production = "powerlaw"
        return super().maker(
            component=production,
            version=version,
            target=target,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )

    def sample(
        self,
        time_start: FloatLike | None = None,
        time_end: FloatLike | None = None,
        diameter_number: PairOfFloats | None = None,
        diameter_number_end: PairOfFloats | None = None,
        diameter_range: PairOfFloats | None = None,
        area: FloatLike | None = None,
        return_age: bool = True,
        **kwargs: Any,
    ) -> np.ndarray:
        """
        Sample diameters and ages from the production function.

        This function can either sample from a given age range or
        from a given cumulative number/diameter pair (but not both).

        Parameters
        ----------
        time_start : FloatLike, optional
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        time_end, FloatLike, optional
            The ending age in units of My relative to the present, which is used to compute the cumulative SFD. The default is 0 (present day).
        diameter_number : PairOfFloats, optional
            A pair of diameter and cumulative number values, in the form of a (D, N). If provided, the function convert this value
            to a corresponding age and use the production function for a given age.
        diameter_number_end : PairOfFloats, optional
            A pair of diameter and cumulative number values, in the form of a (D, N). If provided, the function will convert this
            value to a corresponding reference age and use the production function for a given age.
        diameter_range : PairOfFloats
            The minimum and maximum crater diameter to sample from in meters.
        area : FloatLike, optional
            The area in m² over which the production function is evaluated to generate the expected number, which is the production
            function over the input age/cumulative number range at the minimum diameter.
        return_age : bool, optional
            If True, the function will return the sampled ages in addition to the diameters. The default is True.
        **kwargs: Any
            Any additional keywords that are passed to the function method.

        Returns
        -------
        numpy array
            The sampled diameter values
        numpy array or None
            The sampled age values if return_age is True, otherwise None.

        """
        if "age" in kwargs:
            time_start = kwargs.pop("age")
        arguments = {
            "time_start": time_start,
            "time_end": time_end,
            "diameter_number": diameter_number,
            "diameter_number_end": diameter_number_end,
            "diameter_range": diameter_range,
            "area": area,
            "return_age": return_age,
            **kwargs,
        }
        arguments = self._validate_sample_args(**arguments)
        time_start = arguments["time_start"]
        time_end = arguments["time_end"]
        diameter_range = arguments["diameter_range"]
        area = arguments["area"]
        return_age = arguments["return_age"]
        validate_inputs = kwargs.pop("validate_inputs", False)

        # Build the cumulative distribution function from which we will sample
        input_diameters = np.logspace(np.log10(diameter_range[0]), np.log10(diameter_range[1]))
        cdf = self.function(
            diameter=input_diameters,
            time_start=time_start,
            time_end=time_end,
            validate_inputs=validate_inputs,
            **kwargs,
        )
        if cdf.ndim > 1:
            cdf = cdf[:, 0]
        expected_num = cdf[0] * area if area is not None else None
        diameters = mc.get_random_size(
            diameters=input_diameters,
            cdf=cdf,
            mu=expected_num,
            **vars(self.common_args),
        )
        if diameters.size == 0:
            return np.empty(0), np.empty(0)

        if return_age:
            time_subinterval = np.linspace(time_end, time_start, num=1000)
            N_vs_age = self.function(
                diameter=diameters,
                age=time_subinterval,
                validate_inputs=validate_inputs,
                **kwargs,
            )

            # Normalize the weights for each diameter
            if N_vs_age.ndim > 1:
                N_vs_age -= np.min(N_vs_age, axis=1)[:, None]  # Subtract the min along age dimension
                N_vs_age = N_vs_age[:, :-1]  # Remove the last value to avoid the cumulative sum reaching 1
                weights_sum = np.sum(N_vs_age, axis=1)[:, None]  # Sum of weights for each diameter
                weights = N_vs_age / weights_sum  # Normalize weights for each diameter

                # Compute the CDF for each diameter
                cdf = np.cumsum(weights, axis=1)

            else:
                if len(N_vs_age) > 0:
                    N_vs_age -= np.min(N_vs_age)  # Subtract the min along age dimension
                    N_vs_age = N_vs_age[:-1]  # Remove the last value to avoid the cumulative sum reaching 1
                    weights_sum = np.sum(N_vs_age)  # Sum of weights for each diameter
                    weights = N_vs_age / weights_sum  # Normalize weights for each diameter
                    # Compute the CDF for each diameter
                    cdf = np.cumsum(weights)
                else:
                    cdf = np.empty(0)

            # Generate uniform random numbers for each diameter
            random_values = self.rng.uniform(0, 1, size=diameters.size)
            if N_vs_age.ndim > 1:
                chosen_subinterval_indices = np.zeros(diameters.size, dtype=int)
                # For each diameter, find the corresponding index in the CDF
                for i, cdf_row in enumerate(cdf):
                    # Find the corresponding index in the CDF for each diameter
                    chosen_subinterval_indices[i] = np.searchsorted(cdf_row, random_values[i], side="right") - 1
            else:
                chosen_subinterval_indices = np.searchsorted(cdf, random_values[:], side="right") - 1

            # Ensure indices are within valid range
            chosen_subinterval_indices = np.clip(chosen_subinterval_indices, 0, len(time_subinterval) - 2)
            # Sample a random age within the selected subinterval for each diameter
            time_lo = time_subinterval[chosen_subinterval_indices]
            time_hi = time_subinterval[chosen_subinterval_indices + 1]
            timevals = self.rng.uniform(low=time_lo, high=time_hi)

            # Sort the ages and diameters so that they are in order of decreasing age
            if timevals.size > 1:
                sort_indices = np.argsort(timevals)[::-1]
                diameters = diameters[sort_indices]
                timevals = timevals[sort_indices]
        else:
            timevals = np.empty(0)

        return diameters, timevals

    @abstractmethod
    def function(
        self,
        diameter: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        time_start: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        time_end: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
        **kwargs: Any,
    ) -> NDArray[np.float64]: ...

    def age_from_D_N(
        self,
        diameter: FloatLike | Sequence[FloatLike] | ArrayLike,
        cumulative_number_density: FloatLike | Sequence[FloatLike] | ArrayLike,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> FloatLike | ArrayLike:
        """
        Return the age in My for a given number density of craters and diameter.

        Parameters
        ----------
        diameter : float-like or array-like
            diameter of the crater in m
        cumulative_number_density : float-like or array-like
            number density of craters per m² surface area greater than the input diameter
        validate_inputs : bool, optional
            If True, the function will validate the inputs. The default is True.
        **kwargs: Any
            Any additional keywords that are passed to the function method.

        Returns
        -------
        float_like or numpy array
            The age in My for the given relative number density of craters.
        """
        if validate_inputs:
            diameter, cumulative_number_density = self._validate_csfd(
                diameter=diameter, cumulative_number_density=cumulative_number_density
            )

        def _root_func(t, D, N):
            kwargs.pop("validate_inputs", None)
            retval = self.function(diameter=D, time_start=t, validate_inputs=False, **kwargs) - N
            return retval

        xtol = 1e-8
        x0 = 4000.0
        if self.valid_time[0] is not None:
            xlo = self.valid_time[0]
        else:
            xlo = 0.0
        if self.valid_time[1] is not None:
            xhi = self.valid_time[1]
        else:
            xhi = 1e6
        retval = []
        darr = np.array(diameter)
        narr = np.array(cumulative_number_density)
        converged = []
        flag = []
        for i, d in np.ndenumerate(darr):
            sol = root_scalar(
                lambda x: _root_func(x, d, narr[i]),
                x0=x0,
                xtol=xtol,
                method="brentq",
                bracket=[xlo, xhi],
            )
            retval.append(sol.root)
            converged.append(sol.converged)
            flag.append(sol.flag)
        retval = np.array(retval)
        converged = np.array(converged)
        flag = np.array(flag)
        if np.all(converged):
            return retval.item() if np.isscalar(diameter) else retval
        else:
            raise ValueError(
                f"The root finding algorithm did not converge for all values of diameter and cumulative_number. Flag {flag}"
            )

    def D_from_N_age(
        self,
        cumulative_number_density: FloatLike | Sequence[FloatLike] | ArrayLike,
        time_start: FloatLike | Sequence[FloatLike] | ArrayLike,
        time_end: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> FloatLike | ArrayLike:
        """
        Return the diameter for a given number density of craters and age.

        Parameters
        ----------
        cumulative_number_density : float-like or array-like
            number density of craters per m² surface area greater than the input diameter
        time_start : FloatLike or ArrayLike, default=1.0
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        time_end, FloatLike or ArrayLike, optional
            The ending age in units of My relative to the present, which is used to compute the cumulative SFD. The default is 0 (present day).
        validate_inputs : bool, optional
            If True, the function will validate the inputs. The default is True.
        **kwargs: Any
            Any additional keywords that are passed to the function method.

        Returns
        -------
        float_like or numpy array
            The diameter for the given number density and age.
        """
        if "age" in kwargs:
            time_start = kwargs.pop("age")
        if validate_inputs:
            time_start, time_end = self._validate_age(time_start, time_end)

        def _root_func(D, N, time_start, time_end):
            kwargs.pop("validate_inputs", None)
            retval = (
                self.function(
                    diameter=D,
                    time_start=time_start,
                    time_end=time_end,
                    validate_inputs=False,
                    **kwargs,
                )
                - N
            )
            return retval

        xtol = 1e-8
        x0 = 1.0  # Initial guess for the diameter
        xlo = 1e-6
        xhi = 1e6
        retval = []
        narr = np.array(cumulative_number_density)
        converged = []
        flag = []
        for _, n in np.ndenumerate(narr):
            sol = root_scalar(
                lambda x: _root_func(x, n, time_start, time_end),
                x0=x0,
                xtol=xtol,
                method="brentq",
                bracket=[xlo, xhi],
            )
            retval.append(sol.root)
            converged.append(sol.converged)
            flag.append(sol.flag)
        retval = np.array(retval)
        converged = np.array(converged)
        flag = np.array(flag)
        if np.all(converged):
            return retval.item() if np.isscalar(cumulative_number_density) else retval
        else:
            raise ValueError(
                f"The root finding algorithm did not converge for all values of cumulative_number and age. Flag {flag}"
            )

    @abstractmethod
    def chronology(
        self,
        time_start: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        time_end: FloatLike | Sequence[FloatLike] | ArrayLike = 0.0,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> FloatLike | ArrayLike: ...

    @abstractmethod
    def csfd(self, diameter: FloatLike | ArrayLike, **kwargs: Any) -> FloatLike | ArrayLike: ...

    def _validate_csfd(
        self,
        diameter: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
        cumulative_number_density: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
    ) -> tuple[FloatLike | ArrayLike, FloatLike | ArrayLike]:
        """
        Validates the diameter and cumulative_number arguments.

        Both arguments can be either scalar or array-like, but they must be both scalars or both arrays of the same length.  Values must be non-negative.

        Parameters
        ----------
        diameter : float-like or array-like, optional
            Diameter of the impacts
        cumulative_number_density: float-like or array-like, optional
            Cumulative number of impacts larger than `diameter`

        Raises
        ------
        ValueError
            If any of the conditions on the CSFD are not met.
        """
        if diameter is None and cumulative_number_density is None:
            raise ValueError("Either the 'diameter' or 'cumulative_number_density' must be provided")

        # Convert inputs to numpy arrays for uniform processing
        if diameter is not None:
            diameter_array = np.atleast_1d(diameter)
        else:
            diameter_array = None
        if cumulative_number_density is not None:
            cumulative_number_density_array = np.atleast_1d(cumulative_number_density)
        else:
            cumulative_number_density_array = None

        # Check if both are provided, they should be of the same length
        if diameter_array is not None and cumulative_number_density_array is not None:
            # Check for length consistency
            if len(diameter_array) != len(cumulative_number_density_array):
                raise ValueError("The 'diameter' and 'cumulative_number_density' must have the same length when both are provided")

        # Validate non-negative values
        if diameter_array is not None and np.any(diameter_array < 0):
            raise ValueError("All values in 'diameter' must be non-negative")
        if cumulative_number_density_array is not None and np.any(cumulative_number_density_array < 0):
            raise ValueError("All values in 'cumulative_number_density' must be non-negative")

        if diameter is not None and not np.isscalar(diameter):
            diameter = diameter_array
        if cumulative_number_density is not None and not np.isscalar(cumulative_number_density):
            cumulative_number_density = cumulative_number_density_array
        return diameter, cumulative_number_density

    def _validate_sample_args(self, **kwargs: dict) -> dict:
        """
        Validate all the input arguments to the sample method.

        This function will raise a ValueError if any of the arguments are invalid.  It will also convert age arguments to diameter_number and vice versa.

        Parameters
        ----------
        time_start : FloatLike, optional
            The starting time in units of My relative to the present, which is used to compute the
        time_end, FloatLike, optional
            The ending time in units of My relative to the present, which is used to compute the cumulative SFD.
            The default is 0 (present day).
        diameter_number : PairOfFloats, optional
            A pair of diameter and cumulative number values, in the form of a (D, N), which gives the total cumulative number of
            projectiles, N, larger than diameter, D. If provided, the function convert this value to a corresponding age and use the
            production function for a given age.
        diameter_number_end : PairOfFloats, optional
            A pair of diameter and cumulative number values, in the form of a (D, N), which gives the total cumulative number of
            projectiles, N, larger than diameter, D.. If provided, the function will convert this value to a corresponding time_end
            and use the production function for a given age. The default is (1000.0, 0) (present day).
        diameter_range : PairOfFloats
            The minimum and maximum crater diameter to sample from in meters.
        area : FloatLike, optional
            The area in m² over which the production function is evaluated to generate the expected number, which is the production
            function over the input age/cumulative number range at the minimum diameter.
        return_age : bool, optional
            If True, the function will return the sampled ages in addition to the diameters. The default is True.

        Returns
        -------
        FloatLike
            The start time in units of My relative to the present.
        FloatLike
            The end time in units of My relative to the present.
        PairOfFloats
            A pair of diameter and cumulative number values, in the form of a (D, N) representing the cumulative number and diameter values for a surface at the age given by the start time
        PairOfFloats
            A pair of diameter and cumulative number values, in the form of a (D, N) representing the cumulative number and diameter values for a surface at the age given by the end time.
        PairOfFloats
            The minimum and maximum diameter values to sample from in meters.
        FloatLike
            The area in m² over which the production function is evaluated to generate the expected number, which is the production
            function over the input time/cumulative number range at the minimum diameter.

        Raises
        ------
        ValueError
            If any of the following conditions are met:
            - Neither the age nore the diameter_number argument is provided.
            - Both the age and diameter_number arguments are provided.
            - Both the time_end and diameter_number_end arguments are provided.
            - The time_start or age argument is provided but is not a scalar.
            - The time_end argument is provided but is not a scalar.
            - The diameter_number argument is not a pair of values, or any of them are less than 0
            - The diameter_number_end argument is not a pair of values, or any of them are less than 0
            - The diameter_range is not provided.
            - The diameter_range is not a pair of values.
            - The minimum diameter is less than or equal to 0.
            - The maximum diameter is less than or equal the minimum.
            - The area argument is not a scalar or is less than 0.
        """
        _REF_DIAM = 1000.0  # The diameter value used if age arguments are provided
        age = kwargs.pop("age", None)
        if age is not None:
            kwargs["time_start"] = age
        time_start = kwargs.get("time_start")
        time_end = kwargs.get("time_end")
        diameter_number = kwargs.get("diameter_number")
        diameter_number_end = kwargs.get("diameter_number_end")
        diameter_range = kwargs.get("diameter_range")
        area = kwargs.get("area")
        return_age = kwargs.get("return_age", True)

        if time_start is None and diameter_number is None:
            raise ValueError("Either the 'time_start' or 'diameter_number' must be provided")
        elif time_start is not None and diameter_number is not None:
            raise ValueError("Only one of the 'time_start' or 'diameter_number' arguments can be provided")
        if time_end is not None and diameter_number_end is not None:
            raise ValueError("Only one of the 'time_end' or 'diameter_number_end' arguments can be provided")
        if time_start is not None and not np.isscalar(time_start):
            raise ValueError("The 'time_start' must be a scalar")
        if time_end is not None and not np.isscalar(time_end):
            raise ValueError("The 'time_end' must be a scalar")

        if diameter_range is None:
            raise ValueError("The 'diameter_range' must be provided")
        if len(diameter_range) != 2:
            raise ValueError("The 'diameter_range' must be a pair of values")
        if diameter_range[0] <= 0:
            raise ValueError(f"Diameter range invalid: {diameter_range}. The minimum diameter must be greater than 0")
        if diameter_range[1] < diameter_range[0]:
            raise ValueError(
                f"Diameter range invalid: {diameter_range}. The maximum diameter must be greater than or equal to the minimum diameter"
            )

        if area is None:
            raise ValueError("The 'area' must be provided")
        else:
            if not np.isscalar(area):
                raise ValueError("The 'area' must be a scalar")
            if area < 0.0:
                raise ValueError("The 'area' must be greater than 0")

        if diameter_number is not None:
            if len(diameter_number) != 2:
                raise ValueError("The 'diameter_number' must be a pair of values in the form (D, N)")
            diameter_number = self._validate_csfd(*diameter_number)
            diameter_number_density = (diameter_number[0], diameter_number[1] / area)
            time_start = self.age_from_D_N(*diameter_number_density)
        else:
            diameter_number_density = (
                _REF_DIAM,
                self.function(diameter=_REF_DIAM, age=time_start),
            )
            diameter_number = (
                diameter_number_density[0],
                diameter_number_density[1] * area,
            )

        if time_end is None and diameter_number_end is None:
            diameter_number_end = (_REF_DIAM, 0.0)
            diameter_number_density_end = diameter_number_end
            time_end = 0.0

        if diameter_number_end is not None:
            if len(diameter_number_end) != 2:
                raise ValueError("The 'diameter_number_end' must be a pair of values in the form (D,N)")
            diameter_number_end = self._validate_csfd(*diameter_number_end)
            diameter_number_density_end = (
                diameter_number_end[0],
                diameter_number_end[1] / area,
            )
            # Check to be sure that the diameter in the diameter_number_end is the same as diameter_number.
            # If not, we need to adjust the end diameter value it so that they match
            if diameter_number is not None:
                if diameter_number_density_end[0] != diameter_number_density[0]:
                    diameter_number_density_end = (
                        diameter_number_density[0],
                        self.function(diameter=diameter_number_density[0], time_start=time_end),
                    )
                    diameter_number_end = (
                        diameter_number[0],
                        diameter_number_density_end[1] * area,
                    )

        if time_end is None:
            diameter_number_end = self._validate_csfd(*diameter_number_end)
            diameter_number_density_end = (
                diameter_number_end[0],
                diameter_number_end[1] / area,
            )
            time_end = self.age_from_D_N(*diameter_number_density_end)

        time_start, time_end = self._validate_age(time_start, time_end)

        if diameter_number_end is None:
            diameter_number_end = (
                diameter_number[0],
                self.function(diameter=diameter_number[0], time_start=time_end) * area,
            )

        if not isinstance(return_age, bool):
            raise ValueError("The 'return_age' argument must be a boolean")

        kwargs["time_start"] = time_start
        kwargs["time_end"] = time_end
        kwargs["diameter_number"] = diameter_number
        kwargs["diameter_number_end"] = diameter_number_end
        kwargs["diameter_range"] = diameter_range
        kwargs["area"] = area
        kwargs["return_age"] = return_age

        return kwargs

    def _validate_age(
        self,
        time_start: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        time_end: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
    ) -> FloatLike | ArrayLike:
        """
        Processes the time_start argument and time_end arguments. Checks that they are valid and returns a tuple of age and time_end.

        Parameters
        ----------
        time_start : FloatLike or ArrayLike, default=1.0
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        time_end, FloatLike or ArrayLike, optional
            The reference used when computing age in My. If none is passed, it will be set either 0 or an array of zeros, depending on
            the size of time_start.

        Returns
        -------
        tuple of float
            The start and end ages in units of My.

        Raises
        ------
        ValueError
            If the the start age is greater than the end age or the age variable is not a scalar or a sequence of 2 values.
        """
        if np.isscalar(time_start):
            if not isinstance(time_start, FloatLike):
                raise TypeError("time_start must be a numeric value (float or int)")
            if time_end is None:
                time_end = 0.0
            elif not np.isscalar(time_end) or not isinstance(time_end, FloatLike):
                raise TypeError("time_end must be a numeric value (float or int)")
            if time_start < time_end:
                raise ValueError("time_start must be greater than the time_end")
            if self.valid_time[0] is not None and time_start < self.valid_time[0]:
                raise ValueError(
                    f"time_start must be greater than the minimum valid time {self.valid_time[0]} (it is units of My before present)"
                )
            if self.valid_time[1] is not None and time_start > self.valid_time[1]:
                raise ValueError(
                    f"time_start must be less than the maximum valid time {self.valid_time[1]} (it is in units of My before present)"
                )
        else:
            if isinstance(time_start, (list, tuple)):
                time_start = np.array(time_start, dtype=np.float64)
            elif not isinstance(time_start, np.ndarray):
                raise TypeError("time_start must be a numeric value (float or int) or an array")

            if time_end is None:
                time_end = np.zeros_like(time_start, dtype=np.float64)
            elif np.isscalar(time_end):
                time_end = np.full_like(time_start, time_end, dtype=np.float64)
            elif isinstance(time_end, (list, tuple)):
                time_end = np.array(time_end, dtype=np.float64)
            elif not isinstance(time_end, np.ndarray):
                raise TypeError("time_end must be a numeric value (float or int) or an array")

            if time_start.size != time_end.size:
                raise ValueError("The 'time_start' and 'time_end' arguments must be the same size if both are provided")

            if np.any(time_start < time_end):
                raise ValueError("time_start must be greater than the time_end")

            if self.valid_time[0] is not None:
                if np.any(time_start < self.valid_time[0]) or np.any(time_end < self.valid_time[0]):
                    raise ValueError(
                        f"time_start must be greater than the minimum valid age {self.valid_time[0]} (it is in units of My before present)"
                    )
            if self.valid_time[1] is not None:
                if np.any(time_start > self.valid_time[1]) or np.any(time_end > self.valid_time[1]):
                    raise ValueError(
                        f"time_start must be less than the maximum valid age {self.valid_time[1]} (it is in units of My before present)"
                    )

        return time_start, time_end

    @parameter
    def generator_type(self):
        """
        The type of generator to use.

        This can be either "crater" or "projectile".  This determines the nature of the production function, differentiating between 'crater' and 'projectile' types, which affect the calculations and interpretations of the production function.
        """
        return self._generator_type

    @generator_type.setter
    def generator_type(self, value):
        if not value:
            self._generator_type = self._valid_generator_types[0]
            return
        if not isinstance(value, str):
            raise TypeError("generator_type must be a string")
        if value.lower() not in self._valid_generator_types:
            raise ValueError(f"Invalid generator_type {value}. Must be one of {self._valid_generator_types}")
        self._generator_type = value.lower()
        return

    @property
    def valid_time(self) -> tuple[float, float]:
        """
        The range of ages over which the production function is valid. The range is given in My.

        Returns
        -------
        tuple
            The lower and upper bounds of the valid time range in My, or None if not applicable.
        """
        return (None, None)


import_components(__name__, __path__)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.random import Generator

from cratermaker.constants import FloatLike
from cratermaker.core.base import ComponentBase, import_components

if TYPE_CHECKING:
    from cratermaker.components.projectile import Projectile
    from cratermaker.components.target import Target


class Scaling(ComponentBase):
    """
    The abstract base class for all scaling models. It defines the interface for converting between projectile and crater diameters.
    """

    _registry: dict[str, Scaling] = {}

    def __init__(
        self,
        target: Target | str | None = None,
        projectile: Projectile | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        target : Target | str, default="Moon"
            The target body for the impact. Can be a Target object or a string representing the target name.
        projectile : Projectile | str, default="asteroids"
            The projectile model for the impact. Can be an Projectile object or a string representing the projectile name.
        rng : numpy.random.Generator | None
            |rng|
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            |rng_seed|
        rng_state : dict, optional
            |rng_state|
        **kwargs : Any
            |kwargs|
        """
        from cratermaker.components.projectile import Projectile
        from cratermaker.components.target import Target

        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_target", None)
        object.__setattr__(self, "_projectile", None)
        # combine the kwargs with the common_args, giving common_args priority
        kwargs = {**kwargs, **vars(self.common_args)}
        self._target = Target.maker(target, **kwargs)
        self._projectile = Projectile.maker(projectile, target=self.target, **kwargs)

    @classmethod
    def maker(
        cls,
        scaling: str | Scaling | None = None,
        target: Target | str | None = None,
        projectile: Projectile | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ) -> Scaling:
        """
        Initialize a Scaling model with the given name or instance.

        Parameters
        ----------
        scaling : str, Scaling, or None, default=None
            The name of the scaling model to initialize. If None, the default model is used.
        target : Target | str, default="Moon"
            The target body for the impact. Can be a Target object or a string representing the target name.
        projectile : Projectile | str, default="asteroids"
            The projectile model for the impact. Can be an Projectile object or a string representing the projectile name.
        rng : numpy.random.Generator | None
            |rng|
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            |rng_seed|
        rng_state : dict, optional
            |rng_state|
        **kwargs : Any
            |kwargs|

        Returns
        -------
        Scaling
            An instance of the specified scaling model.

        Raises
        ------
        KeyError
            If the specified scaling model name is not found in the registry.
        TypeError
            If the specified scaling model is not a string or a subclass of Scaling.
        """
        if scaling is None:
            scaling = "montecarlo"
        scaling = super().maker(
            component=scaling,
            target=target,
            projectile=projectile,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )
        if projectile is not None:
            scaling.projectile = projectile
        if target is not None:
            scaling.target = target

        return scaling

    @abstractmethod
    def projectile_to_transient(self, **kwargs: Any) -> np.float64: ...
    @abstractmethod
    def transient_to_projectile(self, **kwargs: Any) -> np.float64: ...
    @abstractmethod
    def transient_to_final(self, transient_diameter: FloatLike, **kwargs: Any) -> tuple[np.float64, str]: ...
    @abstractmethod
    def final_to_transient(self, final_diameter: FloatLike, morphology_type: str | None = None, **kwargs: Any) -> np.float64: ...
    @abstractmethod
    def recompute(self, **kwargs: Any) -> None: ...

    def __str__(self) -> str:
        base = super().__str__()
        return f"{base}\nTarget: {self.target._component_name}\nProjectile: {self.projectile._component_name}"

    @property
    def target(self):
        """The Target object associated with this Scaling object."""
        return self._target

    @target.setter
    def target(self, value):
        from cratermaker.components.target import Target

        self._target = Target.maker(value, **vars(self.common_args))
        return

    @property
    def projectile(self):
        """The Projectile object associated with this Scaling object."""
        return self._projectile

    @projectile.setter
    def projectile(self, value):
        from cratermaker.components.projectile import Projectile

        self._projectile = Projectile.maker(value, **vars(self.common_args))
        return

    @property
    def model(self):
        """The name of the scaling model."""
        return self._component_name


import_components(__name__, __path__)



from __future__ import annotations

import importlib
import pkgutil
import shutil
import sys
from abc import ABC
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import numpy as np
import xarray as xr
from numpy.random import BitGenerator, Generator, RandomState, SeedSequence
from numpy.typing import ArrayLike
from tqdm import tqdm

from cratermaker.utils.general_utils import parameter


@dataclass
class CommonArgs:
    """
    Stores the common arguments that are shared among all components of the project.

    This is used to pass these common arguments to the maker functions of each component when they are initialized from other components. For example, the Scaling component will store an instance of Target and Projectile. If defaults of these components are used, then the Scaling.maker will pass the arguments stored here to the Target.maker and Projectile.maker functions when it initializes the Target and Projectile objects. This way, if the user has specified any non-default arguments for the simulation (e.g. a custom simdir or rng_seed), then those arguments will be passed to the Target and Projectile makers when they are initialized from the Scaling maker, ensuring that all components of the simulation are initialized with the same common arguments.
    """

    simdir: Path
    """The root directory of the simulation."""
    rng: Generator | None
    """The random number generat or object."""
    rng_seed: int | None
    """The random number seed."""
    rng_state: dict | None
    """The current state of the random number generator."""
    ask_overwrite: bool = True
    """Flag indicating whether to prompt the user before overwriting a file."""


class CratermakerBase:
    """
    Base class for the Cratermaker project.

    Parameters
    ----------
    simdir : str | Path
        |simdir|
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    ask_overwrite : bool, optional
        If True, prompt the user for confirmation before overwriting any existing files when saving output. Default is True.
    **kwargs : Any
        |kwargs|
    """

    def __init__(
        self,
        simdir: str | Path | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        ask_overwrite: bool | None = None,
        **kwargs,
    ):
        object.__setattr__(self, "_user_defined", set())
        object.__setattr__(self, "_rng", None)
        object.__setattr__(self, "_rng_seed", None)
        object.__setattr__(self, "_rng_state", None)
        object.__setattr__(self, "_simdir", None)
        object.__setattr__(self, "_output_dir_name", None)
        object.__setattr__(self, "_output_file_pattern", [])
        object.__setattr__(self, "_output_file_prefix", None)
        object.__setattr__(self, "_output_file_extension", "nc")
        object.__setattr__(self, "_output_image_file_extension", "png")
        object.__setattr__(self, "_export_dir_name", "export")
        object.__setattr__(self, "_ask_overwrite", None)
        object.__setattr__(self, "_save_actions", [])
        object.__setattr__(self, "_driver_to_extension_map", {})

        self.simdir = simdir
        self.ask_overwrite = ask_overwrite

        self._rng_seed = rng_seed
        self.rng, self.rng_state = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state)

        super().__init__()

    def reset(
        self,
        files_to_remove: list[Path | str] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Reset the component by removing its output files.

        Parameters
        ----------
        files_to_remove : list[Path | str], optional
            If set, this is the list of files that will be removed. If not set, then the removed files will be determined using the component's output_dir and output_file_pattern attributes.
        **kwargs : Any
            |kwargs|
        """
        if files_to_remove is None:
            files_to_remove = self.saved_output_files()
        if files_to_remove and not self._overwrite_check(files_to_remove):
            print("Reset operation cancelled by user.")
            return

    def to_config(self, remove_common_args: bool = False, **kwargs: Any) -> dict[str, Any]:
        """
        Converts values to types that can be used in yaml.safe_dump. This will convert various types into a format that can be saved in a human-readable YAML file.

        Parameters
        ----------
        obj : Any
            The object whose attributes will be stored.  It must have a _user_defined attribute.
        remove_common_args : bool, optional
            If True, remove the set of common arguments that are shared among all components of the project from the configuration. Default is False.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        dict[str, Any]
            A dictionary of the object's attributes that can be serialized to YAML.

        Notes
        -----
        - The function will ignore any attributes that are not serializable to human-readable YAML. Therefore, it will ignore anything that cannot be converted into a str, int, float, or bool.
        - The function will convert Numpy types to their native Python types.
        """
        return _to_config(self, remove_common_args=remove_common_args, **kwargs)

    def save(
        self,
        interval: int | None = None,
        filename: str | Path | None = None,
        skip_actions: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Runs the action methods specified in the save_actions property of this component, which should be a dictionary where the keys are the names of actions that can be performed on this component (e.g. "plot") when calling the save function and the values are the arguments that should be passed to that action when it is called by the save function.

        Parameters
        ----------
        interval : int or None, optional
            The interval number to save. If None, save without an interval number. Default is None.
        filename : str or Path or None, optional
            The base filename to save to, without the output directory or file extension. If None, the filename will be generated using the output_file_prefix and output_file_extension properties. Default is None.
        skip_actions : bool, optional
            If True, skip running the actions specified in the save_actions property and just return without doing anything. Default is False.
        **kwargs : Any
            |kwargs|

        Notes
        -----
        - skip_actions should be set to True if the save method is called from inside another method of the same object that could potentially be included in a save_actions entry. This will prevent recursive calls to save.
        """
        if skip_actions:
            return

        if filename is None:
            filename = self.output_filename(interval=interval)

        for entry in self.save_actions:
            for action, action_kwargs in entry.items():
                if hasattr(self, action):
                    action_method = getattr(self, action)
                    if callable(action_method):
                        args = {**action_kwargs, **kwargs, "interval": interval}
                        action_method(**args)
                    else:
                        raise ValueError(f"{action} is not a valid action for {self.name}")
                else:
                    raise ValueError(f"{action} is not a valid action for {self.name}")
        return

    def export(self, **kwargs: Any) -> None:
        """
        Export the component data to the specified format.

        This is a stub that acts as a pass-through for components that don't have an export method defined.

        Parameters
        ----------
        **kwargs : Any
            |kwargs|
        """
        pass

    def from_file(self, filename: str | Path, **kwargs: Any) -> None:
        """
        Load the component data from a file.

        This is a stub that acts as a pass-through for components that don't have a from_file method defined.

        Parameters
        ----------
        filename : str or Path
            The path to the file to load the component data from.
        **kwargs : Any
            |kwargs|
        """
        pass

    def saved_output_files(self, **kwargs: Any) -> list[Path]:
        """
        Check if the component has any output files in its output directory.

        Returns
        -------
        list[Path]
            A list of Path objects representing the files that would be removed during a reset operation. Returns an empty list if no files found
        """
        if self.output_dir is None or not self.output_file_pattern:
            return []
        if not self.output_dir.exists():
            return []
        output_file_list = []
        for pattern in self.output_file_pattern:
            output_file_list.extend(self.output_dir.glob(pattern))

        # Add surface image files to the list
        output_file_list.extend(list(self.plot_dir.glob(f"*.{self.output_image_file_extension}")))

        if output_file_list:
            return output_file_list
        else:
            return []

    def output_filename(self, interval: int | None = None, **kwargs: Any) -> str:
        """
        Generate the base output filename for a given interval, but does not include the output directory.

        Parameters
        ----------
        interval : int or None, optional
            The interval number to generate the filename for. If None, the filename will not include an interval number. Default is None.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        str
            The generated output filename for the given interval.
        """
        if self.output_file_prefix is None or self.output_file_extension is None:
            raise ValueError("output_file_prefix and output_file_extension must be set to generate an output filename")
        if interval is not None:
            return f"{self.output_file_prefix}{interval:06d}.{self.output_file_extension}"
        else:
            return f"{self.output_file_prefix}.{self.output_file_extension}"

    def _output_file_processor(self, data_file_list, interval_index: int | None = None, **kwargs: Any) -> xr.Dataset:
        """
        File processor for the output files generated by this component.

        The base method assumes that the output files are in a format that can be read by xarray.open_dataset or xarray.open_mfdataset. Subclasses can override this method to implement custom file processing logic.

        Parameters
        ----------
        interval_index : int or None, optional
            The index of the interval being processed. This can be used to customize the file processor based on the interval. Default is None.
        **kwargs : Any
            |kwargs|
        """
        if interval_index is not None:
            with xr.open_dataset(data_file_list[interval_index]) as ds:
                ds.load()
        elif len(data_file_list) == 1:
            with xr.open_dataset(data_file_list[0]) as ds:
                ds.load()
        else:
            try:
                ds = xr.open_mfdataset(data_file_list, parallel=False, engine="netcdf4", compat="no_conflicts", join="outer")
            except Exception:
                ds = {}  # Fall back when the files cannot be opened together (e.g., because they have different variables or dimensions). In this case, we will open each file separately and concatenate them into a list of Datasets.
                for data_file in tqdm(
                    data_file_list, desc="Reading in files....", unit="files", total=len(data_file_list), position=0, leave=False
                ):
                    with xr.open_mfdataset(data_file, engine="netcdf4") as ds_single:
                        ds[ds_single.interval.item()] = ds_single
        if hasattr(ds, "close"):
            ds.close()
        return ds

    def _overwrite_check(self, files_to_remove: Path | list[Path] | str | list[str]) -> bool:
        """
        Check if the user wants to overwrite existing files or directories, and if they do either from a prompt or based on the `ask_overwrite` attribute, then delete them.

        The user will be prompted to confirm that they want to overwrite existing files or directories if the `ask_overwrite` attribute is set to True. If the user confirms that they want to overwrite the files, then the specified files or directories will be deleted. If the user cancels the operation, then the files will not be deleted and the function will return False. Selecting 'a' will suppress prompts about overwriting files for this operation, but not for future ones.

        Parameters
        ----------
        files_to_remove : Path or list of Path or str or list of str
            The file or list of files that will be removed.

        Returns
        -------
        bool
            True if the user confirmed that they wanted to overwrite the files and that all requested files and/or directories were removed. False if they cancel the operation.
        """
        ask_overwrite = self.ask_overwrite  # This will allow us to temporarily disable prompts about overwriting files for this operation if the user selects 'a' to suppress prompts about overwriting files, without changing the value of self.ask_overwrite for future operations.
        if not isinstance(files_to_remove, list):
            files_to_remove = [files_to_remove]
        for file in files_to_remove:
            file = Path(file)
            if file.exists() and ask_overwrite:
                response = input(
                    f"File '{str(file)}' already exists. To disable this message, set ask_overwrite=False to this instance. Overwrite? (y/N/a): "
                )
                if response.lower() == "a":
                    ask_overwrite = False
                elif response.lower() != "y":
                    print("Operation cancelled by user.")
                    return False
            try:
                if file.is_dir():
                    shutil.rmtree(file)
                else:
                    file.unlink(missing_ok=True)
            except Exception as e:
                print(f"Error removing file {file}: {e}")
                return False
        return True

    def add_save_action(self, action: dict[str, dict]) -> None:
        """
        Add an action to the save_actions property of this component.

        Parameters
        ----------
        action : dict[str, dict]
            A dictionary where the keys are the names of actions that can be performed on this component (e.g. "plot") when calling the save function and the values are the arguments that should be passed to that action when it is called by the save function.
        """
        if not isinstance(action, dict):
            raise TypeError(
                "action must be a dictionary containing a key with a valid action for this component (e.g. 'plot') and the values are the arguments"
            )
        for action_name, args in action.items():
            if hasattr(self, action_name):
                action_method = getattr(self, action_name)
                if not callable(action_method):
                    raise ValueError(f"{action_name} is not a valid action for {self.name}")
            else:
                raise ValueError(f"{action_name} is not a valid action for {self.name}")
            if not isinstance(args, dict):
                raise TypeError(
                    f"Arguments for action {action_name} must be a dictionary of keyword arguments to be passed to the {self.name}.{action_name}() method."
                )
        self._save_actions.append(action)
        return

    @parameter
    def save_actions(self) -> list[dict[str, dict]]:
        """
        A dictionary where the keys are the names of actions that can be performed on this component (e.g. "plot") when calling the save function and the values are the arguments that should be passed to that action when it is called by the save function.
        """
        return self._save_actions

    @save_actions.setter
    def save_actions(self, value) -> None:
        if not isinstance(value, list):
            raise TypeError(
                "save_actions must be a list where each entry is a dictionary containing a key with a valid action for this component (e.g. 'plot') and the values are the arguments"
            )
        for entry in value:
            if not isinstance(entry, dict):
                raise TypeError(
                    "save_actions must be a list where each entry is a dictionary containing a key with a valid action for this component (e.g. 'plot') and the values are the arguments"
                )
            for action, args in entry.items():
                if hasattr(self, action):
                    action_method = getattr(self, action)
                    if not callable(action_method):
                        raise ValueError(f"{action} is not a valid action for {self.name}")
                else:
                    raise ValueError(f"{action} is not a valid action for {self.name}")
                if not isinstance(args, dict):
                    raise TypeError(
                        f"Arguments for action {action} must be a dictionary of keyword arguments to be passed to the {self.name}.{action}() method."
                    )
        self._save_actions = value

    @parameter
    def ask_overwrite(self):
        """
        Flag to indicate whether the user should be prompted to overwrite any old files or not.
        """
        return self._ask_overwrite

    @ask_overwrite.setter
    def ask_overwrite(self, value):
        if value is None:
            value = True
        if not isinstance(value, bool):
            raise TypeError("ask_overwrite must be a bool")
        self._ask_overwrite = value

    def read_saved_output(self, interval: int | None = None, **kwargs) -> tuple[xr.Dataset, ...] | None:
        """
        Read the saved output files for this component and return them as xarray Datasets.

        Parameters
        ----------
        interval : int or None, optional
            The interval number to read. If None, read all intervals. Default is None.
        **kwargs : Any
            |kwargs|

        Returns
        -------
            tuple[xr.Dataset, ...] or None
                A tuple of xarray Datasets containing the data from the saved output files. The order of the Datasets in the tuple corresponds to the order of the file patterns returned by the output_file_pattern property. If no output files are found, returns None.
        """
        import re

        if len(self.output_file_pattern) == 0 or self.output_dir is None:
            return None
        output_ds = []

        for pattern in self._output_file_pattern:
            data_file_list = list(Path(self.output_dir).glob(pattern))
            if len(data_file_list) == 0:
                output_ds.append(None)
                continue
            interval_numbers = []
            if len(data_file_list) > 1:
                for data_file in data_file_list:
                    n = data_file.name.split(self.output_file_prefix)[-1].split(f".{self.output_file_extension}")[0]
                    interval_numbers.append(int(n))
            else:  # handle the special case where the output file pattern does not include an interval number (e.g. "grid.nc")
                n = re.findall(r"\d+", data_file_list[0].name)
                if n:
                    interval_numbers.append(int(n[-1]))

            if len(interval_numbers) > 1:
                tup = sorted(zip(data_file_list, interval_numbers, strict=True))
                data_file_list, interval_numbers = zip(*tup, strict=True)
            data_file_list = list(data_file_list)

            # map the requested interval to the index of interval_numbers
            if interval is not None and len(interval_numbers) > 0:
                if interval < 0:
                    interval_index = interval
                elif interval in interval_numbers:
                    interval_index = interval_numbers.index(interval)
                else:
                    output_ds.append(xr.Dataset())
                    continue
            else:
                interval_index = None

            output_ds.append(self._output_file_processor(data_file_list=data_file_list, interval_index=interval_index, **kwargs))

        if len(output_ds) == 1:
            return output_ds[0]
        else:
            return tuple(output_ds)

    @parameter
    def simdir(self):
        """
        |simdir|

        Returns
        -------
        Path
            The initialized simulation directory as a Path object. Will be a relative path if possible, otherwise will be absolute. If it doesn't exist, it will be created.
        """
        return self._simdir

    @simdir.setter
    def simdir(self, value):
        if isinstance(value, Path):
            self._simdir = value
        elif isinstance(value, (str | None)):
            self._simdir = _simdir_init(value)

    @property
    def output_dir(self) -> Path | None:
        """
        The output directory for a component. If None, the component does not have an output directory set.
        """
        if self._output_dir_name is None:
            return None
        output_dir = self.simdir / self._output_dir_name
        if not output_dir.exists():
            try:
                output_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                raise RuntimeError(f"Could not create output directory at {output_dir}") from e
        return output_dir

    @property
    def export_dir(self) -> Path | None:
        """
        The export directory for a component. If None, the component does not have an export directory set.
        """
        if self._export_dir_name is None:
            return None
        export_dir = self.simdir / self._export_dir_name
        if not export_dir.exists():
            try:
                export_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                raise RuntimeError(f"Could not create export directory at {export_dir}") from e
        return export_dir

    @property
    def plot_dir(self) -> Path:
        """
        The directory for plots.
        """
        if self._output_dir_name is None:
            return None
        plot_dir = self.simdir / f"{self._output_dir_name}_images"
        if not plot_dir.exists():
            try:
                plot_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                raise RuntimeError(f"Could not create output directory at {plot_dir}") from e
        return plot_dir

    @property
    def output_image_file_extension(self) -> str:
        """
        The file extension to use for output images.
        """
        return self._output_image_file_extension

    @output_image_file_extension.setter
    def output_image_file_extension(self, value: str):
        """
        Set the file extension to use for output images.

        Parameters
        ----------
        value : str
            The file extension to use for output images (e.g., "png", "jpg", "tif").
        """
        self._output_image_file_extension = value

    @property
    def output_file_pattern(self) -> list[str]:
        """
        Return a list of one or more file patterns that this component will generate when saved.

        This is used when reading save output files or when restting the components. Some components may generate multiple output files with different patterns (e.g. the observed and emplaced crater files generated by the counting component), so this should return a list of all patterns that are generated by this component. If the component does not generate any files, this should return an empty list.

        Returns
        -------
        list[str]
            A list of file patterns (e.g., ["surface*.nc"]) that this component will generate when saved.
            If the component does not generate any files, return an empty list.
        """
        return self._output_file_pattern

    @property
    def output_file_prefix(self) -> str | None:
        """
        Return the file prefix for this component's output files. This is used to identify the files generated by this component when reading saved output or resetting the component. If the component does not generate any files, this should return None.

        Returns
        -------
        str or None
            The file prefix (e.g., "surface") for this component's output files, or None if the component does not generate any files.
        """
        return self._output_file_prefix

    @property
    def output_file_extension(self) -> str | None:
        """
        Return the file extension for this component's output files. This is used to identify the files generated by this component when reading saved output or resetting the component. If the component does not generate any files, this should return None.

        Returns
        -------
        str or None
            The file extension (e.g., "nc") for this component's output files, or None if the component does not generate any files.
        """
        return self._output_file_extension

    @parameter
    def rng_seed(self):
        """
        The random rng_seed for the simulation RNG.

        Returns
        -------
        int or None
            The integer rng_seed used to initialize the RNG, or None if not set.
        """
        return self._rng_seed

    @rng_seed.setter
    def rng_seed(self, value):
        if value is not None:
            if not isinstance(value, (int | np.integer)) or np.isnan(value) or np.isinf(value) or value < 0:
                raise TypeError("rng_seed must be a positive integer")
            self._rng_seed = int(value)
        else:
            self._rng_seed = None

    @property
    def rng(self):
        """
        The random number generator used for stochastic elements of the simulation.

        Returns
        -------
        numpy.random.Generator or None
            The RNG instance, or None if not initialized.
        """
        return self._rng

    @rng.setter
    def rng(self, value):
        self._rng, _ = _rng_init(rng=value, rng_seed=self.rng_seed, rng_state=self.rng_state)

    @parameter
    def rng_state(self):
        """
        The state of the random number generator.

        Returns
        -------
        dict or None
            A dictionary representing the RNG state, or None if the RNG is not initialized.
        """
        return self.rng.bit_generator.state if self.rng is not None else None

    @rng_state.setter
    def rng_state(self, value):
        _, self._rng_state = _rng_init(rng=self.rng, rng_seed=self.rng_seed, rng_state=value)

    @property
    def common_args(self) -> CommonArgs:
        return CommonArgs(
            simdir=self.simdir,
            rng=self.rng,
            rng_seed=self.rng_seed,
            rng_state=self.rng_state,
            ask_overwrite=self.ask_overwrite,
        )

    @property
    def driver_to_extension_map(self) -> dict[str, str]:
        """A dictionary mapping valid export/from_file drivers for this component to the corresponding file extensions that can be used when exporting files with that driver."""
        return self._driver_to_extension_map

    @property
    def valid_drivers(self) -> list[str]:
        """A list of valid export/from_file drivers for this component."""
        return list(self._driver_to_extension_map.keys())

    @property
    def valid_extensions(self) -> list[str]:
        """A list of valid export/from_file file extensions for this component."""
        return list(self._driver_to_extension_map.values())


class ComponentBase(CratermakerBase, ABC):
    """
    Base class for components of the Cratermaker project.

    Defines the common parameters and methods for all components in the Cratermaker project, including the maker class that is used to select the correct component from user arguments.


    """

    _registry: dict[str, type[ComponentBase]] = {}

    def __init__(self, **kwargs: Any) -> None:
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        **kwargs : Any
            |kwargs|
        """
        object.__setattr__(self, "_save_actions", {})
        super().__init__(**kwargs)

    def __str__(self) -> str:
        # Return just the name of the class
        base_class = type(self).__mro__[1].__name__
        return f"<{base_class}: {self._component_name}>"

    @classmethod
    def maker(
        cls,
        component: str | type[ComponentBase] | ComponentBase | None = None,
        **kwargs: Any,
    ) -> ComponentBase:
        """
        Initialize a component model with the given name or instance.

        Parameters
        ----------
        component : str or ComponentBase or None
            The name of the component to use, or an instance of ComponentBase. If None, it choose a default component.
        kwargs : Any
            |kwargs|

        Returns
        -------
        component
            An instance of the specified component model.

        Raises
        ------
        KeyError
            If the specified component model name is not found in the registry.
        TypeError
            If the specified component model is not a string or a subclass of component.
        """
        if component is None:
            component = cls.available()[0]  # Default to the first available component
        if isinstance(component, str):
            if component not in cls.available():
                raise KeyError(f"Unknown component model: {component}. Available models: {cls.available()}")
            return cls._registry[component](**kwargs)
        elif isinstance(component, type) and issubclass(component, ComponentBase):
            return component(**kwargs)
        elif isinstance(component, ComponentBase):
            return component
        else:
            raise TypeError(f"component must be a string or a subclass of component, not {type(component)}")

    @parameter
    def name(self):
        """
        The registered name of this scaling model set by the @ComponentBase.register decorator.
        """
        return self._component_name

    @classmethod
    def register(cls, name: str):
        """
        Class decorator to register a component model component under the given key.
        """

        def decorator(subcls):
            subcls._component_name = name
            subcls._registry[name] = subcls
            return subcls

        return decorator

    @classmethod
    def available(cls) -> list[str]:
        """Return list of all registered catalogue names."""
        return list(cls._registry.keys())


def import_components(package_name: str, package_path: list[str]) -> None:
    """
    Import all modules of a component package, optionally skipping private modules.

    Parameters
    ----------
    package_name : str
        The full name of the package (e.g., "cratermaker.components.component").
    package_path : list[str]
        The __path__ attribute of the package (usually just __path__).
    """
    for _, module_name, _ in pkgutil.iter_modules(package_path):
        importlib.import_module(f"{package_name}.{module_name}")


def _rng_init(
    rng: Generator | None = None,
    rng_seed: int | ArrayLike | SeedSequence | BitGenerator | Generator | RandomState | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> tuple[Generator, dict]:
    """
    Initialize the random number generator (RNG) based on the provided rng_seed.

    Parameters
    ----------
    rng : Generator, optional
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        Set the rng_state of the RNG.

    Returns
    -------
    Generator
        The initialized RNG instance.
    dict
        The state of the RNG.

    Notes
    -----
    - If rng is provided, it will be used directly and any input rng_seed or rng_state will be ignored.
    - If both rng_seed and rng_state are provided, rng_state will take precedence.
    - If neither rng nor rng_seed is provided, a new RNG will be created using the default rng_seed and the state will be returned.
    """
    if rng is not None:
        if not isinstance(rng, Generator):
            raise ValueError("rng must be a numpy.random.Generator instance.")
    else:
        if rng_seed is None:
            rng = np.random.default_rng()
        if rng_state is not None:
            if not isinstance(rng_state, dict):
                raise ValueError("rng_state must be a dictionary.")
            try:
                rng.bit_generator.state = rng_state
            except Exception as e:
                raise ValueError("Invalid rng_state provided.") from e
        elif rng_seed is not None:
            try:
                rng = np.random.default_rng(seed=rng_seed)
            except Exception as e:
                raise ValueError("Invalid rng_seed provided.") from e

    rng_state = rng.bit_generator.state
    return rng, rng_state


def _simdir_init(simdir: str | Path | None = None, **kwargs: Any) -> Path:
    """
    Initialize the simulation directory.

    Parameters
    ----------
    simdir : str | Path | None
        |simdir|

    Returns
    -------
    Path
        The initialized simulation directory as a Path object. Will be a relative path if possible, otherwise will be absolute.
    """
    if simdir is None:
        p = Path.cwd()
    else:
        try:
            p = Path(simdir)
            if not p.is_absolute():
                p = Path.cwd() / p
            p.mkdir(parents=True, exist_ok=True)
            p = p.resolve()
        except TypeError as e:
            raise TypeError("simdir must be a path-like object (str, Path, or None)") from e
    try:
        simdir = p.relative_to(Path.cwd())
    except ValueError:
        simdir = p
    return simdir


def _convert_for_yaml(obj):
    if isinstance(obj, dict):
        return {k: _convert_for_yaml(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_convert_for_yaml(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(_convert_for_yaml(v) for v in obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, (np.integer, np.floating)):
        return obj.item()
    elif isinstance(obj, Path):
        return str(obj)
    elif isinstance(obj, (str, int, float, bool)):
        return obj
    elif obj is None:
        return None
    else:
        return str(obj)


def _to_config(obj, remove_common_args: bool = False, **kwargs: Any) -> dict[str, Any]:
    config = _convert_for_yaml({name: getattr(obj, name) for name in obj._user_defined if hasattr(obj, name)})
    if remove_common_args:
        config = {key: value for key, value in config.items() if key not in obj.common_args.__dict__}
    return {key: value for key, value in config.items() if value is not None}



import json
import math
from pathlib import Path
from typing import Any

import numpy as np
from cratermaker._cratermaker import counting_bindings
from numpy.typing import NDArray
from scipy import fft

from cratermaker.components.counting import _MIN_FACE_FOR_COUNTING, _N_LAYER, Counting
from cratermaker.components.crater import Crater
from cratermaker.components.surface import LocalSurface, Surface


@Counting.register("simplecount")
class SimpleCount(Counting):
    """
    A basic crater counting model that uses depth-to-diameter values to estimate degradation states of craters for counting.

    Parameters
    ----------
    surface : Surface | LocalSurface
        The surface or local surface view to be counted.
    **kwargs : Any
        |kwargs|
    """

    def __init__(self, surface, **kwargs: Any):
        super().__init__(surface=surface, **kwargs)
        self._component_name = "simplecount"

    def measure_degradation_state(self, crater: Crater, **kwargs: Any) -> float:
        """
        Measure the degradation state of a crater by using a variation of the depth-to-diameter proxy for degradation state from Minton et al. (2019) [#]_ and Riedel et al. (2020) [#]_.

        Parameters
        ----------
        crater : Crater
            The crater to measure.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        float
           The estimated degradation state of the crater in m². The crater object will also have its degradation state value updated in place

        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021
        .. [#] Riedel, C., Minton, D.A., Michael, G., Orgel, C., Bogert, C.H. van der, Hiesinger, H., 2020. Degradation of Small Simple and Large Complex Lunar Craters: Not a Simple Scale Dependence. Journal of Geophysical Research: Planets 125, e2019JE006273. https://doi.org/10.1029/2019JE006273


        Notes
        -----
        The technique used here is a major improvement over the one used previously in CTEM. Rather than using the depth-to-diameter directly, it instead uses the ratio of the measured depth-to-diameter to the initial computed depth-to-diameter of the crater. Fitting functions for both simple and complex craters were developed using this metric.
        """
        from cratermaker.constants import _VSMALL

        # A simplified transition diameter for the two morphology regimes. A more comprehensive study including other planetary bodies is needed to improve this
        _DTR = 10.0e3

        # The following are empirically-derived constants determined through numerical experiments
        c0 = 0.2771649498066961
        c1 = 0.1032775007149389
        d_scale = 178.40182304735092
        s0 = 0.11511000581387373
        s1 = 0.017078961587273955
        p0 = 2.0700104291102317
        p1 = 0.0920770504148557

        def a_vs_diameter(diameter):
            if diameter < _DTR:
                return s0 + s1 * np.log(diameter)
            else:
                dkm = diameter * 1e-3
                dtrkm = _DTR * 1e-3
                return c0 + c1 * (1.0 - np.exp((dtrkm - dkm) / d_scale))

        def p_vs_diameter(diameter):
            if diameter > _DTR:  # In the fit, we fixed the exponent p to be constant for complex craters
                diameter = _DTR
            return p0 + p1 * np.log(diameter)

        def K_vs_depth_over_depth_orig(depth_over_depth_orig, a, p):
            return a * (depth_over_depth_orig ** (-1 / p) - 1.0) * crater.measured_radius**2

        if crater.measured_depth_to_diameter is None:
            return 0.0
        depth_over_depth_orig = crater.measured_depth_to_diameter / crater.depth_to_diameter
        if depth_over_depth_orig < _VSMALL:
            depth_over_depth_orig = _VSMALL
        a = a_vs_diameter(crater.measured_diameter)
        p = p_vs_diameter(crater.measured_diameter)
        crater.degradation_state = K_vs_depth_over_depth_orig(depth_over_depth_orig, a, p)

        return crater.degradation_state

    def visibility_function(self, crater: Crater, Kv1: float = 0.30, gamma: float = 2.0, **kwargs: Any) -> float:
        """
        Calculate the visibility function for a crater using eq. 7 from Minton et al. (2019) [#]_.

        Parameters
        ----------
        crater : Crater
            The crater to calculate the visibility function for.
        Kv1: float
            The visibility function parameter Kv1 from Minton et al. (2019). Default value is 0.30 based on recalibration with Cratermaker.
        gamma: float
            The visibility function parameter gamma from Minton et al. (2019). Default value is 2.0.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        float
            The visibility function value for the crater in units of m².

        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021

        """
        return Kv1 * crater.measured_radius**gamma
